/*;=============================================================================*/
/* ;	<make docs> includes C comments beginning w/ "/*;" or "//;"		*/
/*;	Ting 90, Muench, Rideau, Hohensee 99; .hpr'01				*/
/*; hp:	too many changes to be mentioned here, re <./d/eForth.S.orig>		*/
/*;=============================================================================*/
/*;*/
/*;*/
/*; stack notation:*/
/*; name	( in -- out )		headerflags*/
/*; 	word description*/
/*;*/
/*; header flags:*/
/*;	C: use only while compiling,	I: immediatly executing*/
/*;	F: word leaves <eflags> according to opr, for -re- "flg"*/
/*;*/
/*;*/
.list
.psize 0
.att_syntax noprefix

/* +++++++++++++++++++++++++++++++++++++++++++*/
/* ; host, OS or program type specific*/
// ; customize some lo-level alternatives:

// ; linux specific; using "fork" if undefined
with_vfork = 0
#ifdef __VFORK
.ifc __VFORK,Linux
with_vfork = 1	// ; by Makefile, use "vfork"
.endif
#endif

// ;set to non-zero for multi-byte chars
_mbchar_= 1
#ifdef __MBCHAR	
_mbchar_=__MBCHAR	// ; passed by value(!) from Makefile
#endif

// ;if set, implement "DO..LOOP" instead of "FOR..NEXT", adds about 692 bytes to binary
_doloop_= 0
#ifdef __DOLOOP
_doloop_= -1		// ; by Makefile
#endif

// ; also implement "FOR..NEXT"
_fornx_ = 0
#ifdef __FORNXT
_fornx_ = -1
#endif

// ; use redirected signal handlers
_sighd_ = 0
#ifdef __SIGHD
_sighd_ = 1		// ; by Makefile
#endif

// ; initial, default radix for numeric interpretation
BASE0	= 10

// ; more configureable assignments, below.

/* +++++++++++++++++++++++++++++++++++++++++++*/

ORIGIN	= 0x08048000	// ; Linux/ELF program base address

#define __INITSEG__ .text 0 // ; initial code segment
#define __DATASEG__ .data 0 // ; data segment, before code
#define __CODESEG__ .data 1 // ; code segment, after data

.include "scfctl.inc"	// ; file & i/o ctrls
.include "scdefs.inc"	// ; syscall names, macro

.set PAGE_SIZE,(1<<page_shift)
.set PAGE_MASK,-(PAGE_SIZE)
#define pg(x) (((x)+PAGE_SIZE-1)&PAGE_MASK)

.set SIGHUP,1
.set SIGQUIT,3
.set SIGILL,4
.set SIGBUS,7
.set SIGFPE,8
.set SIGSEGV,11
.set SIGCHLD,17
.set SIGWINCH,28
.set SA_SIGINFO,4	// ;asm/signal.h
.set SA_RESETHAND,0x80000000

stdin	= 0
stdout	= 1
stderr	= 2

BSP	=	8	// ; backspace
HT	=	9	// ; horizontal tabulator
DEL	=	127	// ; delete
LF	=	10	// ; line feed, Linux cr/lf, 'new line' = <nl>
// ;CR	=	13	// ; carriage return
// ;ESC	=	27	// ; escape
BL	=	32	// ; blank space
// ;TIC	=	0x27	// ; single quote

/* +++++++++++++++++++++++++++++++++++++++++++*/

/* ; Version control	*/

// ;startup <relse.sh> w/	#updt: item
// ;REL	=	2		# r+ :Major release version
// ;VER	=	00		# v+ :Minor release version
// ;EXT	=	00		# x+ :Major extension
// ;TAG	=	00		# t+ :Minor extension
// ;VERNUM =	0x${REL}${VER}${EXT}${TAG} # r.v.x.t :version no.
#include "../i/version.inc"

/* +++++++++++++++++++++++++++++++++++++++++++*/
/* ; forth */

CPLO	=	0x40		// ; Lexicon compile-only bit
IMED	=	0x80		// ;  ...immediate

MASKK	=	0xfFfFfF1F	// ; Lexicon mask - [hp] enabling full size character set
pchmask =	0xffa07e20	// ; hi|lo ranges of printing chars
OCALL	=	0xE8909090	// ; CALL opcode, padded to dword boundary
		# should be padded with a 1 cycle 3-byte NOP.
		#	- [hp] why not the proper jump label?! -
OCALL	=	0xe800408d	// ; lea 0(eax),eax; call...*/

CELSZ	=	4		# Size of a cell
CELLMASK =	-CELSZ		# cell size alignment mask
NBPW	=	32		# bits per word (cell)

VOCSS	=	8		# Depth of vocabulary stack
		#	for greater VOCSS adjust user vari access!
PADSZ	=	512		# disp from "here" to "pad"
QRYSZ	=	PADSZ/2+4	# max chars by "query"

//;
/* +++++++++++++++++++++++++++++++++++++++++++*/
//; memory map
//;	[.text: startup code][.data: data & kernel]...[.bss: stacks, tib, uvari]
//;
//;		size:		useage:
//;RS_SZ	  64K		# return-stack
//;DS_SZ	  64K		# data-stack
//;US_SZ	   4K		# user-vari
//;NM_SZ	  64K		# word-headers
//;CP_SZ	 128K		# compile-space
//;TB_SZ	   4K		# terminal input buffer
//;
//;EM		DSTBT+DS_SZ+CP_SZ # disp to top of initial *virtual* memory
//;
//;		-ve disp wrt EM:
//;NAMTP	0		# top of word-header space, top down
//;RSTBT	NAMTP+NM_SZ	# Top of return stack, top down
//;TIBB		RSTBT+RS_SZ+TB_SZ # Terminal input buffer (TIB), bottom up
//;UPTR		TIBB+US_SZ	# user area (UP0), bottom up
//;DSTBT	UPTR+PAGE_SIZE	# Top of data stack, top down
//;
//;RPTR		RSTBT+6*CELSZ	# orig. return stack pointer (RP0), on different L1 cache line
//;SPTR		DSTBT+2*CELSZ	# orig. data stack pointer (SP0)
//;
//;
//;
.set RS_SZ,65536	//; 64K
.set DS_SZ,65536	//; 64K
.set US_SZ,1024*4	//;  4K
.set NM_SZ,65536	//; 64K
.set CP_SZ,65536*2	//;128K
.set TB_SZ,1024*4	//;  4K

.set NAMTP,0		
.set RSTBT,pg(NAMTP+NM_SZ)
.set TIBB,pg(RSTBT+RS_SZ+TB_SZ)
.set UPTR,pg(TIBB+US_SZ)
.set DSTBT,pg(UPTR+PAGE_SIZE)

.set RPTR,RSTBT+6*CELSZ	
.set SPTR,DSTBT+2*CELSZ	

.set EM,pg(DSTBT+DS_SZ+CP_SZ+PAGE_SIZE-1)

/* +++++++++++++++++++++++++++++++++++++++++++*/
/* ;variables*/

/*;*/
//; internal, register useage (mmx/fp regs not used):
//;    eax low div/mod, aux	RL
//;    ebx top on data-stack	TS
//;    ecx word address		WA
//;    edx high div/mod, aux	RH
//;    ebp return-stack		RS
//;    esi instruction ptr	IP
//;    edi user ptr		US
//;    esp data-stack		PS
#define IP	esi
#define PS	esp
#define RS	ebp
#define	TS	ebx
#define US	edi
// ; forth vm
#define RL	eax
#define RH	edx
#define WA	ecx

_USER	=	4*CELSZ		/* ;first user variable offset*/

/* +++++++++++++++++++++++++++++++++++++++++++*/
/* ;macroes*/

/* ;translate from other sources*/
#define	dd	.long
#define dw	.word
#define db	.byte
#define DD	dd
#define DW	dw
#define DB	db
.macro resb n
.rept
	.byte 0
.endr
.endm
#define RESB	resb

#define d_align	.align CELSZ,0
#define c_align	.align CELSZ,0x90

#define c_pad3	lea 0(eax),eax	// ;pad code by 3 bytes

#define lit(x) DOLIT,x

/* ;parameter-stack access:*/

#define TOS(x)	(((x)-1)*CELSZ)(PS)
#define TOS0	TS
#define TOS1	(PS)

/* ;#arg times pop TS, no flags affected
   ; - pop, and lea if num>1, "swap drop" if num=0*/
.macro spop num=1
.if \num > 2
.list
	movl (\num-1)*4(PS),TS
	leal (\num)*4(PS),PS
.nolist
.else
    .if \num > 1
    .list
	popl TS
    .nolist
    .endif
    .if \num > 0
    .list
	popl TS
    .nolist
    .else
	leal 4(PS),PS	//;; swap drop
    .endif
.endif
.endm

/* ;#arg times push TS, no flags affected*/
.macro spush num=1
.rept \num
.list
	pushl TS
.nolist
.endr
.endm

/* ;return-stack access:*/

#define TOR(x)	((x)*CELSZ)(RS)
#define TOR0	(RS)
#define TOR1	4(RS)

/* ;pop from RS to arg, dft=TS, no flags affected
   ; "rdrop" if arg = 0 */
.macro rpop reg=TS
.list
.ifnc \reg,0
	movl (RS),\reg
.endif
	leal 4(RS),RS
.nolist
.endm

/* ;push arg, dft=TS, to RS, no flags affected*/
.macro rpush reg=TS
.list
	leal -4(RS),RS
	movl \reg,(RS)
.nolist
.endm

/* ;disps into pushal/popal regs storage block*/
s_tor	=	32
s_eax	=	28
s_ecx	=	24
s_edx	=	20
s_ebx	=	16
s_esp	=	12
s_ebp	=	8
s_esi	=	4
#define s_edi

#define s_IP	s_esi
#define s_PS	s_esp
#define s_RS	s_ebp
#define	s_TS	s_ebx
#define s_US	s_edi
#define s_RL	s_eax
#define s_RH	s_edx
#define s_WA	s_ecx

/* ;read from "pushad" block, no flags affected*/
.macro readad to,from
.list
    movl s_\from(esp),\to
.nolist
.endm

/* ;write to "pushad" block, no flags affected*/
.macro writead to,from
.list
    movl \from,s_\to(esp)
.nolist
.endm

/* +++++++++++++++++++++++++++++++++++++++++++*/
/* ;word headers*/

/* ;inline direct threading.*/
#define __ENTER leal CELSZ(IP),IP
#define __NEXT  jmpl *(IP)

// ;primitive - headerless if name-field empty
.macro __CODE lex=0,name,label,prev=0,mod=0
.list
	__CODESEG__ 
	c_align
.ifeq \mod
\label:
.endif
.ifnes "\name",""
	__DATASEG__
	d_align
	.long \label
    .ifndef n0
n0:	.long 0
    .else
	.long 10001b
    .endif
10001:
n\label:
	.byte (10000f-1-.)|(\lex&-32)
	.ascii "\name"
10000:
_NAME    = n\label
	d_align
_NTOP	=.
	__CODESEG__
.endif
.ifeq \mod
j\label:
.endif
.nolist
.endm

// ;hi-level, colon defn
.macro __COLON lex=1,name,label,prev=0
.list
	__CODE \lex,"\name",\label,\prev,1
	c_pad3
\label:
	call DOLST
j\label:
.nolist
.endm

// ; kernel space user vari
.macro __USER lex=1,name,label,prev=0
.list
.if \lex
	__CODE \lex,"\name",\label,\prev,1
\label:
.equ u\label,_USER
	spush
	__ENTER
	leal u\label(US),TS
	__NEXT
.endif
	_USER = _USER+CELSZ
.nolist
.endm

/* ;kernel space variable*//*saves no memory, is faster*/
.macro __VARI lex=1,name,label,prev=0
.list
	__CODE \lex,"\name",\label,\prev,1
\label:
	spush
	__ENTER
	movl $v\label,TS
	__NEXT
	c_align
v\label:
.nolist
.endm

/*; ****************  Main entry points and COLD start data *******************/

	__INITSEG__

	d_align
	.globl _start

// ; for the "sh" word
SHT:	.asciz "/bin/sh"
	d_align
RG1:.asciz "sh"
	d_align
RG2:.asciz "-c"
	d_align
ARGS:	dd RG1,RG2,CMD+4,0
	d_align

.include "scdata.inc"

ttyt:	.asciz "/dev/tty"
self0t:	.asciz "/proc/self/fd/0"

	d_align
ttyst:	/* ;defaults for "stty sane"*/
	.long 0x2102
	.long 0x0005
	.long 0x04bf
	.long 0x8A3B
	.byte 0
l1:
	.byte 0x03
	.byte 0x1C
	.byte 0x7F
	.byte 0x15
	.byte 0x4
	.byte 0x0
	.byte 0x1
	.byte 0x0
	.byte 0x11
	.byte 0x13
	.byte 0x1A
	.byte 0x0
	.byte 0x12
	.byte 0x0F
	.byte 0x17
	.byte 0x16
	.byte 0x0
l2:
	.rept NCCS-17 /* ;ttyst_disp,(l2-l1),(2b-1b), etc..., - any comment - ? */
		      /* ;gAS ": Error: bad or irreducible absolute expression; zero assumed"*/
	.byte 0
	.endr

	d_align

.if _sighd_
//
/* Singnals Handling */
// ;
.set NSIG,64
.set SIGRTMIN,32
// ;da ist was :falsch:, die auswahl spielt nicht -re- sigchld, wrt "shell"
// ; signo by bit-no.
// ;		 10987654321098765432109876543210
.set fpesig,   0b00000000000000000000000100000000	// ; use special or, parent's hnd
.set winchsig, 0b00010000000000000000000000000000
.set segvsig,  0b00000000000000000000100010000000	// ; segv & buserror
.set ilegsig,  0b00000000000000000000000000010000
.set chldsig,  0b00000000000000100000000000000000	// ; silently return, noop
.set quitsig,  0b00000000000000000000000000001000	// ; ctr/"\"
// ;		 UpIwPVxXuTtTsCcStApUsUkFbAtIqIh
// ;		 NwOiRTfCrTtStOhTeLiSeSiPuBrLuNu
// ;		 Ur nOAsPgOiToNlKrRpRgRlEsRaLiTp
// ;		 S  cFLzU UnPpTdFmMe2v1l  Tp t  
// ;		 E  h R         L
// ;		 D    M         T
// ;	disabled:
.set fpesig,	0	/*; 164 b - 15108,14944 */
.set winchsig,	0	/*;  64 b - 15172,15108 */
//;.set segvsig,	0
//;.set ilegsig,	0
.set quitsig,	0	/*;  52 b - 14892,14944 */
.set chldsg,	0
sa_x:	// ; signals which exit
	.long b
	.long SA_SIGINFO// ;;flags
	.long 0		// ;; restorer (_is_ used!)
	.long 0		// ;; mask
	.long 0		// ;; (extended)
sa_m:	// ; signals which send a message
sa_i:	// ; rt sigaction
	.long r
	.long SA_SIGINFO
	.long 0
	.long 0
	.long 0
.if winchsig
sa_winch:// ; winch is 
	.long winch
	.long SA_SIGINFO
	.long 0
	.long 0
	.long 0
.endif
.if fpesig
sa_fpe:	// ; fpe exits
	.long fpe
	.long SA_SIGINFO
	.long 0
	.long 0
	.long 0
.endif
.ifeq segvsig|ilegsig
.else
//    .ifeq ilegsig
//    .else
sa_segv:// ;
	.long segv	
	.long SA_SIGINFO
	.long 0
	.long 0
	.long 0
//    .endif
.endif
.if quitsig
sa_quit:
	.long quits
	.long SA_SIGINFO
	.long 0
	.long 0
	.long 0
.endif
.if chldsig
sa_chld:
	.long c
	.long SA_SIGINFO
	.long 0
	.long 0
	.long 0
.endif
// ;- signals which restore terminal & exit
// ;		; 3         2         1         0
// ;		;10987654321098765432109876543210
.set sgnsys,   0b01010101000100100010000100101000	// ; use special or, parent's hnd
.set sgnsys,sgnsys&~(fpesig|winchsig|segvsig|ilegsig|quitsig|chldsig)
// ;		$fedcba9876543210fedcba9876543210
.set sgmask,   0b01000001011000011000101011010110	// ; exitting, else message
.set sgmask,sgmask|fpesig|winchsig|segvsig|ilegsig|quitsig|chldsig
	c_align
isegf:
//;     jmp 9f
	pushal
	movl $0,edi
	movl edi,edx
	leal 31(edi),ebx	// ; available signal no's, 1..31
	leal NSIG/8(edi),esi
1:
	movl $sgnsys,eax	// ; signals which remain unchanged
	btl ebx,eax
	jc 2f			// ;?; rt-signal, only
	movl $sa_x,ecx		// ; handler for exit sgs
	movl $sgmask,eax	// ; signals which exit
	btl ebx,eax
	jc 3f			// ;?; sig-exit
	addl $(sa_i-sa_x),ecx	// ; messageing signals
3:
	sys_ rt_sigaction
// ; rt signals
2:
	addl $SIGRTMIN,ebx
	movl $sa_m,ecx		// ; rt signal hnd
	sys_ rt_sigaction
	subl $SIGRTMIN+1,ebx
	jns 1b
// ; specific
.if fpesig
	movl $SIGFPE,ebx
	movl $sa_fpe,ecx
	sys_ rt_sigaction
.endif
.if winchsig
	movl $SIGWINCH,ebx
	movl $sa_winch,ecx
	sys_ rt_sigaction 
.endif
.if ilegsig
	movl $SIGBUS,ebx
	movl $sa_segv,ecx
	sys_ rt_sigaction
	movl $SIGILL,ebx
	movl $sa_segv,ecx
	sys_ rt_sigaction
.endif
.if segvsig
	mov $SIGSEGV,ebx
	movl $sa_segv,ecx
	sys_ rt_sigaction 
.endif
.if quitsig
	movl $SIGQUIT,ebx
	movl $sa_quit,ecx
	sys_ rt_sigaction
.endif
.if chldsig
	movl $SIGCHLD,ebx
	movl $sa_chld,ecx
	sys_ rt_sigaction
.endif
	popal
9:
	ret
// ; ! 1st level subroutine !
// ; regs unchanged, eflags=Z if signal from <hupjobs>, else NZ
// ; store signo in (eof) for <bbi> if @(eof) was < 2
// ; if compiled from source w. __AU_DBG swich %define-d, 
// ;	display signal no. & location of occurrence
sigr:
	cmpl $1,uIERR(US)
	pushl edx
	pushl ecx
	pushl eax
	leal 20(esp),edx
	movl (edx),ecx
	movl 8(edx),edx		// ; arg(2)
	movl 19*4(edx),edx	// ; posn(19) -> pc at offending opr, pc
	jg 1f
	call sigp
	negl ecx
	pushal
	pushl edx
	movl ecx,eax
	movl $(('$'*256+':')*256+'G')*256+'S',edx	/*"SG:$"*/
	call e4pn.s
	movl (esp),eax
	movl $(':'*256+'t')*256+'a',edx			/*"at:"*/
	call e4pn.l
	popl edx
	leal -ORIGIN(edx),eax
	movl $((':'*256+'h')*256+'t')*256+'4',edx	/*"4th:"*/
	call e4pn.s
	popal
1:
///;	cmpl $hupjobs.k,edx
///;	jz 2f			// ;?; prevent false signo from internal call
///;	movl ecx,uIERR(US)
2:
	popl eax
	popl ecx
	popl edx
	ret
sigp:
	pushal
	sys_ getpid
	movl $(('$'*256+':')*256+'P')*256+LF,edx	/*;cr,"P:$"*/
	call e4pn.s
	popal
	ret
// ; print (un)signed sedecimal eax, chars(edx)
e4pn.l:
	xorl ebx,ebx
	jmp 1f
e4pn.s:// ;	;<signed)
	movl eax,ebx
	sarl $24,ebx
	movb $8,bl
	jns 1f
	negl eax
1:
	testl edx,edx
	jz 0f
	pushal
	movl $4,edx
	leal s_edx(esp),ecx
	movl $stderr,ebx
	sys_ write
	popal
0:
e4pn.n:
	leal -12(esp),esp
	movl $8,ecx
	movb $' ',(esp)
2:
	movb $15,dl
	andb al,dl
	cmpb $10,dl
	jc 3f
	addb $7,dl
3:
	addb $'0',dl
	movb dl,(esp,ecx)
	decb bl
	js 4f
	testl eax,eax
	jz 5f
4:
	shrl $4,eax
	decl ecx
	jg 2b
5:
	movl $9,edx
	leal (esp,ecx),ecx
	testb bl,bl
	js 6f
	testl ebx,ebx
	movzbl bl,ebx
	jns 7f
	decl ecx
	incl edx
	movb $'-',(ecx)
7:
	incl ebx
	subl ebx,edx
6:
	movb $' ',(ecx,edx)
	incl edx
	movl $stderr,ebx
	sys_ write
	leal 12(esp),esp
	ret
hupjobs:
	pushal
	sys_ getgid	// ; process group
	movl eax,ebx
	movl $SIGHUP,ecx
	negl ebx
	sys_ kill	// ; remove tasks but not caller - NOTE: false implementation in Linux/Mandrake!
hupjobs.k:
	popal
	ret

.if fpesig
fpe:// ; special entries for idiv ebx, div ebx, 
// ; others should supply entry to hnd on stack and remove immediately after div opr.
// ; on stack supplied handler should continue in dividing 4th word after div opr.
	pushal
	leal s_tor+4(esp),eax	// ; tor
	movl 8(eax),eax		// ; arg(2)
	leal 19*4(eax),edx	// ; posn(19) -> pc at offending opr
	movl $fpebx,ecx		// ; div error hnd, div by @ecx
	movl (edx),eax
	movl (eax),eax
	cmpb $0xf3,ah		// ; div
	jz 1f			// ;?; ebx
	movl $fpebxi,ecx	// ; hnd idiv by @ebx
	cmpb $0xfb,ah		// ; idiv
	jz 1f
	movl $2f,ecx		// ; ret into @esp supplied handler
1:
	movl ecx,(edx)		// ; sto handler
	popal
2:
	ret
.endif
.if segvsig
segv:
	incl (sgf)
	cmpl $1,(sgf)
	jg b
	call sigr
	pushl eax
	movl 12(esp),eax	// ; arg(2)
	movl $cold,19*4(eax)	// ; posn(19) -> pc at offending opr
	popl eax
	ret
.endif
.if quitsig
// ; restart by <CTR"\"> key
quits:
	pushl eax
	movl 12(esp),eax	// ; arg(2)	;;12;
	movl $warm,19*4(eax)	// ; posn(19) -> pc
	popl eax
	ret
.endif
//
// ;- (display signo & pc,) pass signo to bye & exit.
b:
	call sigr
	mov (esp),ebx
	testl ebx,ebx
	jns 1f
	negl ebx
1:
	jmp xBBI
// ;- (display signo & pc,) continue main program.
r:
	call sigr
	nop
c:// ; sigchld
	ret
.if fpesig
fpebx:// ; div 0, div ovf
	cmpl $1,ebx
	jz rpureg	// ;?; ovf from div 1
	jc rzreg	// ;?; div 0
	pushl eax	// ; dvd.hi
	movl edx,eax	// ; dvd.hi
	xorl edx,edx
	divl ebx
	popl eax	// ; dvd.lo
	divl ebx	// ; leave rmd
	ret
fpebxi:// ; idiv 0, idiv ovf
	cmpl $-1,ebx
	jz rmureg	// ;?; ovf from div -1
	jl 1f		// ;?; other -ve
	cmpl $1,ebx
	jz rpureg	// ;?; ovf from div 1
	jl rzreg	// ;?; div0
1:
	pushl eax	// ; dvd.hi
	movl edx,eax	// ; dvd.hi
	xorl edx,edx
	idivl ebx
fpebxi.l:
	movl eax,edx
	popl eax	// ; dvd.lo
	idivl ebx	// ; leave rmd
	ret
ruureg:
	incl edx
	jz rpureg
	decl edx
	jz rpureg.r
	negl edx
	mov edx,eax
rmureg:
	negl eax
rpureg:
	movl $0,edx
rpureg.r:
	ret
rzreg:
	orl edx,eax
	jz rzreg.e	// ;?; 0/0
	sarl $31,edx
	notl edx	// ; +ve:-1 -ve:0
	leal 0x80000001(,edx,2),eax// ;q: +/- max
	mov eax,edx
	ret
rzreg.e:
	incl eax
	ret
.endif
.if winchsig
// ; SIGWINCH handler, updates main console window sizes for {max-xy} - n.i. -
winch:	movl eax,esp
	andl $-4096,esp
	popl ebx
	popl ecx
	popl edx
	popl esi
	popl edi
	movl IP,uWCH(US)	//:TEST: - cannot use US! corrupted*/
	popl ebp
	xchg (esp),eax
	movl esp,(esp)
.endif
.else//; _sighd_
.equ sigp,return
.endif//; _sighd_
return:
	ret

/*;*/
/*; ************************* Forth Entry *******************************/
/*;*/

	c_align
_start:
	popl ecx		/*; argc in %ecx, argv is %esp*/
	leal 4(esp,ecx,4),eax	/*; env*/
// ;	movl ecx,argc4e		/// { env @ arg @ - 4/ 2- } is arguments count
	movl esp,(argv4e)
	movl eax,(env4e)
	xorl ebx,ebx
	sys_ brk
	movl eax,(obssi)
	leal EM(eax),ebx
	andl $PAGE_MASK,ebx
	sys_ brk		//; set top of .bss range
	movl eax,(ontop)
	movl eax,(brkaddr)
	movl $0xfeededda,TS	//;eax: initialize TOS
	leal -SPTR(eax),PS	//;esp:
	leal -RPTR(eax),RS	//;ebp:
	leal -TIBB(eax),WA	//;ecx: terminal input buffer
	leal -UPTR(eax),US	//;edi:
	movl PS,(os0)
	movl RS,(or0)
	movl WA,(otib)
	movl US,(ou0)
	movl US,(up2)
	spush
.if _sighd_
	call isegf
.endif
	movl $COLD1,IP		//;esi: to high level cold start
	__NEXT

/* ;COLD start moves the following to USER variables.*/
/* ;MUST BE IN SAME ORDER AS USER VARIABLES.*/
	__DATASEG__
	d_align

// ; "+origin"
/* ;location of "+origin", at beginning of writable data-mem*/
porg:	.long cold
	.long warm
obssi:	.long 0		/*;initial base of code/data memory*/
UZERO:	/*; initial uvari, should be kept in sequence! */
ou0:	.long	0,0,0,0	/*;reserved for m-tasking*/
argc4e:	.long	0	/*; -reserve- */
argv4e:	.long	0
env4e:	.long	0
brkaddr:.long	0
os0:	.long	0	/*;SP0*/
or0:	.long	0	/*;RP0*/
otqkey:	.long	QRX	/*;?KEY*/
otemit:	.long	TXSTO	/*;EMIT*/
otexp:	.long	ACCEP	/*;EXPECT*/
ottap:	.long	KTAP	/*;TAP*/
otecho:	.long	TXSTO	/*;ECHO*/
otprompt:.long	DOTOK	/*;PROMPT*/
obase:	.long	BASE0	/*;BASE*/
otmp:	.long	0	/*;tmp*/
ospan:	.long	0	/*;SPAN*/
otoin:	.long	0	/*;>IN*/
ontib:	.long	0	/*;#TIB*/
otib:	.long	0	/*;TIB*/
oblk:	.long	0	/*;BLK */
ocsp:	.long	0	/*;CSP*/
oteval:	.long	INTER	/*;EVAL*/
otnum:	.long	NUMBQ	/*;NUMBER*/
ohld:	.long	0	/*;HLD*/
ohnd:	.long	0	/*;HANDLER*/
ocntx:	.long	0	/*;CONTEXT pointer*/
.rept VOCSS
	.long	0	/*;vocabulary stack*/
.endr
ocurnt:	.long	0	/*;CURRENT pointer*/
ovocl:	.long FORTHv	/*;voc-link*/
ocp:	.long CTOP	/*;CP*/
ontop:	.long 0		/*;NP*/
olast:	.long LASTN	/*;LAST*/
okbd:	.long -1	/*;KBD & flag*/
ostdin:	.long stdin
ostdou:	.long stdout
ostder:	.long stderr
oioer:	.long 0		/*;i/o err*/
odocol:	.long DOCOL	/*;BDOCOL (:)*/
odexit:	.long DEXIT	/*;BDEXIT (;)*/
ostate:	.long 0		/*;STATE*/
ULAST:

up2:	.long 0
sgf:	.long 0

stattr://; saved tty attributes
.rept termios_size
	.long -1
.endr

tattr://; modified tty attributes
.rept termios_size
	.long -1
.endr

	__CODESEG__

// ; das spielt:
// ;.equ l,warm-cold
// ;    dd l
// ;.equ utab,1024-(ULAST-UZERO)
// ;    dd utab

/*;*/
/*; ************************* OS-dependent Forth  *******************************/
/*;*/

warm:
	movl (up2),US
	movl uRZERO(US),RS
.if 0 ; -?-
// ; 11 cells put to tos by sighnd if in interpreter loop, no extra data otherwise.
// ; -> leave PS as is, ioctl/select console-i/o to not reponding to sigquit -?-
	cmpl PS,IP
	jc 1f
// ; depends on no. of fd-s(?) and/or sys_select args, by "w4i": +11 cells for left off watching mode bit-table.
	leal 11*CELSZ(PS),PS
1:
	cmpl uSZERO(US),PS
	jc 2f
.endif
	movl uSZERO(US),PS
2:
	movl $jABORT,esi
	__NEXT
cold:
        movl $COLD1,IP	// ; initialize IP
	__NEXT
/*;*/
//; tic		( -- n )
//;	ret l.s.32-bit of cpu clock counter
	__CODE,"tic",TI
	spush
	__ENTER
	rdtsc
	movl eax,TS
	__NEXT
/*;*/
//; bp		( -- )
//;	insert a lo-level breakpoint
	__CODE,"bp",BP
	int $3
	__ENTER
	__NEXT
/*;*/
//; [s]  	( -- a )
//;	syscalls names and numbers list, structure:
//;	{	.long max-syscall,list-size,0
//;	{	[sys]:
//;	repetitive entries
//;	{		.word syscall-number
//;	{		.byte name-length
//;	{		.asciz syscall-name ("__NR_" stripped)
//;	list ends w/	.long 0
	__CODE,"[s]",BSYSB
	pushl TS
	__ENTER
	movl $syscalref,TS
	__NEXT
/*;*/
//; (s)		( a -- n | er )*/
//;	fetch syscall-no. n from list @[sys] by byte-counted name at a.
//;	-ve a: display all names, ret n := 0.
	__CODE,"(s)",PSYSP
	movl TS,eax
	testl TS,TS
	__ENTER
	pushl edi		// ;save reg to PS
	pushl esi		// ;
	pushl $0		// ;prepare error flag
.set p_US,8
	mov $syscalref-2,edx	// ;names, ptr to pre-initial item stg posn <.byte 0,0 :<>: 0,0>
	mov -10(edx),ecx	// ;no. of syscalls
	js 1f
	jz 3f
	leal 4(TS),TS		// ;stg after leading 3 chars
5:
	movl -4(TS),eax	// ;count & leading 3 chars, 1st test
1:
	decl ecx
	js 3f			// ;?;ret name not found
	movl edx,edi		// ;list, ptr to stg
	movzbl (edx),edx	// ;stg countbyte
	leal 4(edi,edx),edi	// ;for 16bit no, countbyte, <nul>-terminator
	movl edi,edx		// ;save next stg
	scasl
	jnz 4f			// ;?;not found or, "display all" mode
	movzbl al,eax		// ;stg count
	subl $3,eax
	movl TS,esi
	js 5b			// ;?;..(which shouldn't happen, shortest name is 3 chars)..
	jz 2f			// ;?;done
	xchg eax,ecx
	repz cmpsb
	mov eax,ecx		// ;rst syscalls count
	jnz 5b			// ;?;no match
2:
	movzbl -2(edx),ecx	// ; syscall number
	movl ecx,(esp)
3:
	popl TS
	popl esi
	popl edi
	__NEXT
4:
	testl eax,eax
	jns 1b			// ;?;not found
	cmpl eax,TS
	jnz 1b			// ;?;not found
	pushal
	movl (s_US+p_US)(esp),US
	leal 1(edx),ecx
	movzbl (edx),edx
	movl uSTDOUT(US),ebx
	sys_ write
	leal uTEMP(US),ecx
	movb $BL,(ecx)
	movl $1,edx
	sys_ write
	popal
	jmp 1b
/*;*/
//; ssc		( p n -- p er )
//;	executes OS specific system call n, return value/error code in er,
//;	syscall parameters from table at p, p pointing to 1st item in list:
//;	p points to 1st item which, may directly be taken from {sp@}.
//;		ebx, at lo addr
//;		ecx, edx, esi, edi, until
//;		ebp, at hi addr	(used w. 2.4.+ linux kernels)
//;	example eForth access, all reg values but eax & sp@ are optional:
//;		( edi esi edx ecx ebx sp@ eax -- edi esi edx ecx ebx eax2 )
//;	where eax2 is result, replacing the passed syscall number from eax 
//;	on entry. for instance, reading a directory entry (fd=file descriptor):
//;		{ fd pad 200 sp@ 141 syscall }
//;	or, <sleep> a while (10s):
//;		{ 0 10 sp@ dup sp@ 162 syscall drop }
//;	wait for n milli-seconds:
//;		: ms 1000000 1000000000 */mod sp@ dup sp@ sys nanosleep 6 ndrop ;
//;	send a signal:
//;		33 sp@ sys geteuid swap drop sp@ sys kill
//;	Further reading:
//;		Linux system calls documentation in manual pages section 2,
//;			ebx corresponds to 1st arg in man-pages description.
//;		asmutils "os_linux.inc" for syscall numbers vs names,
//;			http://www.lxhp.in-berlin.de/lhpsyscal.html, some docs...
	__CODE,"ssc",SC
	movl TS,eax		// ; syscall no
	__ENTER
	movl (esp),ebx	// ; end of arguments list
	pushl ebp
	pushl edi
	pushl esi
	movl ebx,edi
	movl (ebx),ebx
	movl 4(edi),ecx
	movl 8(edi),edx
	movl 12(edi),esi
	movl 20(edi),ebp
	movl 16(edi),edi
        sys_			// ; sys_generic
	popl esi
	popl edi
	popl ebp
	movl eax,TS		// ; syscall return value/ernum
	__NEXT
/*;*/
/*; Device dependent I/O*/
/*;*/

/*;*/
//; bye		( -- )
//;	Exit eForth w/ exit code 0.
	__CODE,"bye",BYE
	xorl TS,TS	/* ;exitcode*/
/*;*/
//; bbi		( +n -- )
//;	Exit eForth with non-negative exit-code n.
	__CODE,"bbi",BBI
xBBI:
	call sttys
	sys_ exit
/*;*/
//; w4f		( fd -- )
//;	wait until input pending to channel fd. range limited to -1 < fd < 32.
	__CODE,"w4f",W4F
	__ENTER
	call wf
	spop
	__NEXT
	c_align
wf:
	cmpl $32,TS
	pushl esi
	pushl edi
	jnc 1f
	movl TS,ecx	// ; chan
	movl esp,esi
	movzbl bh,ebx	// ; 0
	btsl ecx,ebx	// ; in
	pushl ebx	// ; r/w/x flags
	movl esp,ecx	// ; ->readfds, also excep, write is NULL
	movl $0x7fffffff,ebx // ; almost endless...
	pushl ebx	// ; timeval.sec
	pushl ebx
	movl esp,edi	// ;time
	pushl esi	// ;save empty rstack
	xorl edx,edx	// ;writefds
	movl ecx,esi	// ;readfds
	leal 32(edx),ebx
	sys_ _newselect	// ;b:fdmax,c:*r,d:*w,S:*x,D:*time
	movl (esp),esp
1:
	popl edi
	popl esi
	ret
/*;*/
/*; ?rx		( -- c T | F )*/
/*;	Ret input character and true, or a false if no input.*/
/* ;eax:=-(ernum), eflags: Z=eof, GT=error*/
	__CODE,"?rx",QRX
	pushl TS
	__ENTER
	movl uSTDIN(US),ebx	// ; stdin
	pushl $0		// ; Input Buffer
	movl esp,ecx
	movl $1,edx		// ; 1 char
	sys_ read
	negl eax
	js 1f			// ; 0:=eof, -1:=char
	addl $4,esp
	xorl eax,eax		// ; drop char, ret 0
1:
	movl eax,TS
	__NEXT
/*;*/
/*; key		( -- c )*/
/*;	Wait for and return an input character.*/
/*;	uvari "err" set to input error code < 1, 1 if char received.*/
	__CODE 3,"key",KEY
	pushl TS
	__ENTER
	movl uSTDIN(US),ebx
	call wf
// ; ?terminal ( -- c ) keycode or 0
// ; in: -/-
// ;out: eax:=-1/+ve ebx:=key/xx eflags:=S/NS+Z
	lea uIERR(US),ecx	// ; temp. char-buff, ernum
	movl uSTDIN(US),ebx
	movl $1,edx
	sys_ read
	movzbl (ecx),TS
	mov eax,(ecx)		// ; save ernum, Z is <eof>, S is ernum, >0 is char(s) received
	negl eax		// ; -1 on success
	js 2f			// ;?; char received
	xorl TS,TS
2:
	__NEXT
/*;*/
/*; tx!		( c -- )*/
/*;	Send character c to the output device.*/
/* eax:=ernum*/
	__CODE,"tx!", TXSTO
.if _mbchar_/*;multi-byte emit*//*8+,13500,08,36(mbchar)*/
	movl TS,edx	// ; byte-cnt or zero
	bsrl ebx,edx	// ; find top bit to send
	leal 7(edx),edx	// ; round up to 8bit-byte size
	shrl $3,edx	// ; bytewise
	jnz 1f		// ;?; non-zero cnt
	incl edx	// ; at least one byte
1:
.else
	movzbl	%bl,%ebx
	movl $1,%edx	// ; char
.endif
	__ENTER
	pushl ebx
	movl uSTDOUT(US),ebx
	leal (esp),ecx // ; from output buffer
	sys_ write
	popl ecx	// ;drop but, leave eax=ernum
	popl TS
	__NEXT
/*;*/
//; !io		( -- )
//;	Initialize the serial I/O devices.
//;	set stdin to zero-timeout input. open a channel for keyboard input
	__CODE,"!io",STOIO
	cmpl $0,uKBD(US)
	__ENTER
	jns IOMODEn
	call vtsto
	call kbopen
	call vtini
	call inonbl
	__NEXT
	c_align
kbopen:
	pushal
	xorl edx,edx
	movl uKBD(US),ebx
	cmpl edx,ebx
	jns vtini_n		// ;?; if already open set kbd mode
	movl $ttyt,ebx
	movl $O_RDWR,ecx
	sys_ open
	movl eax,uKBD(US)
	mov eax,ebx
	jmp vtini_n
	c_align
// ; it might be required to re-opening stdout & stderr for independent sets of console-flags -?- ;
vtini:	pushal
vtini_n:
	movl $TCGETS,ecx
	movl $tattr,edx				// ; will be modified
	sys_ ioctl
	andl $(~(ICANON|ECHO)),termios_c_lflag(edx)	// ; leave sigs enabled, disable erase/fill processing, echo
	orl $ONLCR,termios_c_oflag(edx)			// ; enabling NL<-CR/NL
	andl $(~(INPCK|ISTRIP|IXON)),termios_c_iflag(edx) // ;8bit chars
	movb $1,termios_c_cc+VTIME(edx)		// ; timo * 1/10 s (if ~ICANON)
	movb $1,termios_c_cc+VMIN(edx)		// ; min no. of chars for a single read opr
	movl $TCSETS,ecx
	sys_ ioctl				// ; set new tty mode
	popal
	ret
// ; set stdin to non-blocking mode
inonbl:
	pushal
	movl uSTDIN(US),ebx
	movl $(O_RDWR|O_NONBLOCK),edx
	movl $F_SETFL,ecx
	sys_ fcntl
	popal
	ret
// ; save initial state of stdin, if console channel
vtsto:
	movl uSTDIN(US),ebx
	movl $TCGETS,ecx
	movl $stattr,edx			// ; saved till exit
	sys_ ioctl
	ret
/*;*/
//; sttys  	( -- )
//;	reset console to initial state
	__CODE,"sttys",STTYS
	__ENTER
	call sttys
	__NEXT
sttys:
	cmpl $0,uKBD(US)
	js 1f
	pushal
.ifdef stattr
	movl $stattr,edx
.else
	movl $ttyst,edx
.endif
	movl uKBD(US),ebx
	movl $TCSETS,ecx
	sys_ ioctl
	movl uSTDIN(US),ebx
	sys_ ioctl
	popal
1:
	ret
/*;*/
//; iomode  	( -- )
//;	fcntl stdin to non-blocking & r/o, stdout to sync & w/o.
//;	NOTE:	channels opened w/ sys_dup (etc.) cannot be set to distinct 
//;		non/blocking mode, will be set to 'sync', as output channel.
	__CODE,"iomode",IOMODE
	__ENTER
IOMODEn:
	pushal
	movl uSTDIN(US),ebx
	movl $F_SETFL,ecx
	movl $(O_RDONLY|O_NONBLOCK),edx
	sys_ fcntl
	movl uSTDOUT(US),ebx
	movl $(O_WRONLY|O_SYNC),edx
	sys_ fcntl
	popal
	__NEXT
// ;:UNSINN:...
// ;specific to Linux/ELF, anyway, thus no "argc" because env delimited by <nul>-pointer...
/* ;*/
/* ; argc   	( -- n )*/
/* ;	argc*/
// ;	__USER,"argc",CARGCC,IOMODE
/*;*/
//; wsz		( -- a )		(n.i.)
//;	ptr to main window sizes, chars m.s.w[ bot, right ]l.s.w
	__USER,"wsz",WCH
//
/*;*/
//; arg   	( -- a )
//;	list of ptrs to commandline arguments, delimited by <nul>-ptr
//;	arguments count is { env @ arg @ - 4/ 2- }
	__USER,"arg",CARGC
/*;*/
/*; env		( -- a )*/
/*;	list of ptrs to environment strings, delimited by <nul>-ptr*/
	__USER,"env",CENVC
/*;*/
//; bss		( -- a )*/
//;	current top of .bss section (by sys_brk)
	__USER,"bss",CBRK
/*;*/
/*;********************** The  OS-independent kernel **************************/
/*; System and user variables*/
/*;*/

/*;*/
/*; dovar  	( -- a )		C*/
/*;	Run time routine for VARIABLE and CREATE.*/
	__CODE CPLO,"dovar",DOVAR
	spush
	leal CELSZ(IP),TS
	rpop %esi
	__NEXT
/*;*/
/*; uv		( -- a )*/
/*;	Pointer to base of the user area.*/
	__CODE,"uv",UP
	spush
	__ENTER
	movl US,TS
	__NEXT
/*;*/
//; +rg	( n -- a )
//;	calculate address of system initation data; f.i.g. "+origin".
	__CODE,"+rg",PORIGIN,UP
	__ENTER
	addl $porg,TS
	__NEXT
/*;*/
/*; douser  	( -- a )		C*/
/*;	Run time routine for user variables.*/
	__CODE CPLO,"douser",DOUSE,PORIGIN
	spush
	__ENTER
	movl (IP),TS
	addl US,TS
	rpop IP		/*;Pop return address, Adjust RP*/
	__NEXT
/*;*/
/*; sp0		( -- a )*/
/*;	Pointer to bottom of the data stack.*/
	__USER,"sp0",SZERO
/*;*/
/*; rp0		( -- a )*/
/*;	Pointer to bottom of the return stack.*/
	__USER,"rp0",RZERO
/*;*/
/*; '?key   	( -- a ) '*/
/*;	Execution vector of ?KEY.*/
	__USER,"'?key",TQKEY
/*;*/
/*; 'emit   	( -- a ) '*/
/*;	Execution vector of EMIT.*/
	__USER,"'emit",TEMIT
/*;*/
/*; 'expect  	( -- a ) '*/
/*;	Execution vector of EXPECT.*/
	__USER,"'expect",TEXPE
/*;*/
/*; 'tap    	( -- a ) '*/
/*;	Execution vector of TAP.*/
	__USER,"'tap",TTAP
/*;*/
/*; 'echo    	( -- a ) '*/
/*;	Execution vector of ECHO.*/
	__USER,"'echo",TECHO
/*;*/
/*; 'prompt  	( -- a ) '*/
/*;	Execution vector of PROMPT.*/
	__USER,"'prompt",TPROM
/*;*/
/*; base    	( -- a )*/
/*;	Storage of the radix base for numeric I/O.*/
	__USER,"base", BASE
/* ;*/
/* ; tmp   	( -- a )		C*/
/* ;	A temporary storage location used in parse and find.*/
	__USER,,TEMP
/*;*/
/*; span   	( -- a )*/
/*;	Hold character count received by EXPECT.*/
	__USER,"span",SPAN
/*;*/
/*; >in	   	( -- a )*/
/*;	Hold the character pointer while parsing input stream.*/
	__USER,">in",TOIN
/*;*/
/*; >in	   	( -- a )*/
/*;	Hold the character pointer while parsing input stream.*/
	__USER,"blk",BLK
/*;*/
/*; #TIB   	( -- a )*/
/*;	Hold the current count in the terminal input buffer.*/
/*;	terminal input buffer address at next cell { a 4+ @ }*/
	__USER,"#tib",NTIB
	__USER,,TIBP		//; tib address
/*;*/
/*; csp		( -- a )*/
/*;	Hold the stack pointer for error checking.*/
	__USER,"csp",CSP
/*;*/
/*; 'eval   	( -- a ) '*/
/*;	Execution vector of EVAL.*/
	__USER,"'eval",TEVAL
/*;*/
/*; 'number	( -- a ) '*/
/*;	Execution vector of NUMBER?.*/
	__USER,"'number",TNUMB
/*;*/
/*; hld		( -- a )*/
/*;	Hold a pointer in building a numeric output string.*/
	__USER,"hld", HLD
/*;*/
/*; handler	( -- a )*/
/*;	Hold the return stack pointer for error handling.*/
	__USER,"handler",HANDL
/*;*/
/*; context	( -- a )*/
/*;	A area to specify vocabulary search order.*/
	__USER,"context",CNTXT
	_USER = _USER+VOCSS*CELSZ/*;vocabulary stack*/
/*;*/
/*; current	( -- a )*/
/*;	Point to the vocabulary to be extended.*/
	__USER,"current",CRRNT
	__USER,,VOCL		/*;voc-link*/
/*;*/
/*; cp		( -- a )*/
/*;	Point to the top of the code dictionary.*/
	__USER,"cp",CP
/*;*/
/*; np		( -- a )*/
/*;	Point to the bottom of the name dictionary.*/
	__USER,"np",NP
/*;*/
/*; last   	( -- a )*/
/*;	Point to the last name in the name dictionary.*/
	__USER,"last",LAST
/*;*/
/*; kbd		( -- a )*/
/*;	keyboard/terminal channel, or -ve*/
/*;	flag whether terminal channel opened*/
	__USER,"kbd",KBD
/*;*/
/*; stdin   	( -- a )*/
/*;	standard input channel*/
	__USER,"stdin",STDIN
/*;*/
/*; stdout   	( -- an )*/
/*;	standard output channel*/
	__USER,"stdout",STDOUT
/*;*/
/*; stderr   	( -- an )*/
/*;	standard error channel*/
	__USER,"stderr",STDERR
/*;*/
/*; ierr   	( -- an )*/
/*;	input count/ernum from last stdin opr*/
	__USER,"err",IERR
/*;*/
/*; (:)		( -- an )*/
/*;	docol, hi-level xec list vector*/
	__USER,"(:)",BDOCOL
/*;*/
/*; (;)		( -- an )*/
/*;	dosemis, hi-level list termination vector*/
	__USER,"(\073)",BDEXIT
/*;*/
/*; (;)		( -- an )*/
/*;	uvari which holds the compiling (=/=0) state flag*/
	__USER,"state",STATE
/*;*/
/*; --*/
/*;*/

/*;*/
/*; dolit   	( -- w )		C*/
/*;	Push inline literal.*/
	__CODE CPLO,"dolit",DOLIT
	pushl	TS
	mov	4(esi),TS
	addl	$8,esi
	__NEXT
/*;*/
/*; docol   	( a -- )		C*/
/*;	Process colon list; vectored to "(:)"*/
	__CODE CPLO,"docol",DOLST
	jmp *uBDOCOL(US)

	__CODE,,DOCOL
	__ENTER
	subl $CELSZ,%ebp	// ; push current PC on forth return stack
	movl %esi,(%ebp)
	popl %esi		// ; new list address from RET address
	__NEXT
/*;*/
/*; exit   	( -- )*/
/*;	Terminate a colon definition; vectored to "(;)"*/
	__CODE,"exit",EXIT
	jmp *uBDEXIT(US)

	__CODE,,DEXIT
	__ENTER
	movl	(%ebp),%esi	// ;return address
	addl	$CELSZ,%ebp	// ;Adjust RP
	__NEXT
/*;*/
/*; (n)   	( -- )			C*/
/*;	Run time code for the single index loop.*/
	__CODE CPLO,"(n)",DONXT
	subl	$1,(%ebp)
	__ENTER
	jc	1f		// ; ?decrement below 0
	movl	(%esi),%esi	// ;no, branch back again
	__NEXT
1:	addl	$CELSZ,%ebp	// ;yes, pop the index
	addl	$CELSZ,%esi	// ;continue past the branch offset
	__NEXT
/*;*/
/*; 0branch	( f -- )		C*/
/*;	Branch if flag is zero.*/
	__CODE CPLO,"0branch",ZBRA
	testl TS,TS
	popl TS		// ;pop TOS
	jz BRAN
ZBRAn:
	addl $2*CELSZ,%esi	// ;Point IP to next cell
	__NEXT
/*;*/
/*; -b   	( f -- )		C*/
/*;	Branch if flag is non-zero.*/
	__CODE CPLO,"-b",NZBRA
	testl TS,TS
	popl TS
	jz	ZBRAn
	/* -continued- */
/*;*/
/*; branch  	( -- )			C*/
/*;	Branch to an inline address.*/
	__CODE CPLO, "branch",BRAN
	movl	4(%esi),%esi
	__NEXT
/*;*/
/*; =b   	( n1 n2 -- )		C*/
/*;	Branch if flag is non-zero.*/
	__CODE CPLO,"=b",EQBRA
	movl TS,ecx
	popl eax
	popl TS
	cmpl eax,ecx
	jnz ZBRAn
	movl 4(%esi),%esi
	__NEXT
/*;*/
/*; execute	( ca -- )*/
/*;	Execute the word at ca.*/
	__CODE,"execute",EXECU
	movl TS,eax
	popl TS
	jmp *eax
/*;*/
/*; !		( w a -- )*/
/*;	Pop the data stack to memory.*/
	__CODE,"!",STORE
	popl	(TS)
	__ENTER
	popl	TS
	__NEXT
/*;*/
/*; @		( a -- w )*/
/*;	Push memory location to the data stack.*/
	__CODE,"@",AT
	__ENTER
	movl	(TS),TS
	__NEXT
/*;*/
/*; c!		( c b -- )*/
/*;	Pop the data stack to byte memory.*/
	__CODE,"c!",CSTOR
	popl	eax
	__ENTER
	movb	al,(TS)
	popl	TS
	__NEXT
/*;*/
/*; c@		( b -- c )*/
/*;	Push byte memory location to the data stack.*/
	__CODE,"c@",CAT
	__ENTER
	movzbl	(TS),TS
	__NEXT
/*;*/
/*; rp@		( -- a )*/
/*;	Push the current RP to the data stack.*/
	__CODE,"rp@",RPAT
	pushl	TS
	__ENTER
	movl	ebp,TS
	__NEXT
/*;*/
/*; rp!		( a -- )(R: xx -- yy )	C*/
/*;	Set the return stack pointer.*/
	__CODE CPLO,"rp!",RPSTO
	movl	TS,ebp
	__ENTER
	popl	TS
	__NEXT
/*;*/
/*; r>		( -- w )(R: w -- )	C*/
/*;	Pop the return stack to the data stack.*/
	__CODE CPLO,"r>",RFROM
	pushl	TS
	__ENTER
	movl	(ebp),TS
	addl	$CELSZ,ebp
	__NEXT
.if _doloop_
//;
//;- ni		( n -- w )
//;-	fetch n-th level LOOP index. { 0 ni } is { i }.
__CODE 2, "ni",NI
	__ENTER
	mov (RS,TS,(2*CELSZ)),TS
	__NEXT
/*;*/
/*; r@		( -- w )*/
/*; i		( -- w )*/
/*;	Copy top of return stack to the data stack.*/
/*;	r@ is also the loop index in FOR/NEXT and DO/LOOP*/
	__CODE,"i",I
	__CODE,"r@",RAT
.else
	__CODE,"r@",RAT
.endif
	pushl	TS
	__ENTER
	movl	(ebp),TS
	__NEXT
/*;*/
/*; >r		( w -- )(R: -- w )	C*/
/*;	Push the data stack to the return stack.*/
	__CODE CPLO,">r",TOR
	subl	$CELSZ,ebp
	__ENTER
	movl	TS,(ebp)
	popl	TS
	__NEXT
/*;*/
/*; >rr		( w -- w )(R -- w )	C*/
/*;	copy the data stack to the return stack.*/
	__CODE CPLO,">rr",TORR
	subl	$CELSZ,ebp
	__ENTER
	movl	TS,(ebp)
	__NEXT
/*;*/
/*; r>drop  	( -- )(R w -- )		C*/
/*;	drop the return stack.*/
	__CODE CPLO,"r>drop",RDROP
	__ENTER
	addl	$CELSZ,ebp
	__NEXT
/*;*/
/*; sp@		( -- a )*/
/*;	Push the current data stack pointer.*/
	__CODE,"sp@",SPAT
	pushl	TS
	__ENTER
	movl	esp,TS
	__NEXT
/*;*/
/*; sp!		( a -- )*/
/*;	Set the data stack pointer.*/
	__CODE 3,"sp!",SPSTO
	movl	TS,esp
	__ENTER
	popl	TS
	__NEXT
/*;*/
//; sp>		( a -- )
//;	clear data stack
	__CODE,"sp>",SPGT
	movl uSZERO(US),esp
	__ENTER
	popl TS
	__NEXT
/*;*/
//; ndrop  	( wn..w1 n -- )
//;	Discard n cell sized items from stack
	__CODE,"ndrop",NDROP
	leal (esp,TS,4),esp
/*;*/
/*; drop   	( w -- )*/
/*;	Discard top stack item.*/
	__CODE,"drop",DROP
	__ENTER
	popl TS
	__NEXT
/*;*/
/*; 2drop  	( w w -- )*/
/*;	Discard two items on stack.*/
	__CODE,"2drop",DDROP
	__ENTER
	spop 2
	__NEXT
/*;*/
/*; dup		( w -- w w )*/
/*;	Duplicate the top stack item.*/
	__CODE,"dup",DUP
	__ENTER
	pushl	TS
	__NEXT
/*;*/
/*; 2dup  	( w1 w2 -- w1 w2 w1 w2 )*/
/*;	Duplicate top two items.*/
	__CODE,"2dup",DDUP
	movl (esp),eax
	__ENTER
	pushl TS
	pushl eax
	__NEXT
/*;*/
/*; swap   	( w1 w2 -- w2 w1 )*/
/*;	Exchange top two stack items.*/
	__CODE,"swap",SWAP
	movl (esp),eax
	__ENTER
	movl TS,(esp)
	movl eax,TS
	__NEXT
/*;*/
/*; over   	( w1 w2 -- w1 w2 w1 )*/
/*;	Copy second stack item to top.*/
	__CODE,"over",OVER
	pushl	TS
	__ENTER
	movl	CELSZ(%esp),TS
	__NEXT
/*;*/
/*; o-   	( w1 w2 -- w1 w3 )	F*/
/*;	subtract second stack item from top, { over - }.*/
	__CODE,"o-",OVERM
	__ENTER
	subl (esp),TS
	__NEXT
/*;*/
/*; 0<		( n -- t )		F*/
/*;	Ret true if n is negative.*/
	__CODE,"0<",ZLESS
	__ENTER
	sarl $31,TS
	__NEXT
/*;*/
/*; and		( w1 w2 -- w3 )		F*/
/*;	Bitwise AND - w3 bits set only if set in w1 AND w2*/
	__CODE,"and",ANDD
	popl	eax
	__ENTER
	andl	eax,TS
	__NEXT
/*;*/
/*; or		( w1 w2 -- w3 )		F*/
/*;	Bitwise OR - w3 bits set if set in any of w1 OR w2.*/
	__CODE,"or",ORR
	popl	eax
	__ENTER
	orl	eax,TS
	__NEXT
/*;*/
//; xor		( w1 w2 -- w3 )		F
//;	Bitwise exclusive OR - w3 bits set only if not equal in w1 and w2
	__CODE,"xor",XORR
	popl eax
	__ENTER
	xorl eax,TS
	__NEXT
/*;*/
//; um+		( u u -- sum cy )*/
//;	Add two unsigned single numbers and return sum and carry.
	__CODE,"um+",UPLUS
	addl TS,TOS(1)
	sbbl TS,TS
	__ENTER
	negl TS
	__NEXT
/*;*/
//; pick   	( wn...w0 +n -- wn...w0 wn )
//; pick   	( wn...w0 w1 -n -- )
//;	Copy (+n)th stack item to tos.
//;	Copy w1 to (-n)th stack item.
	__CODE,"pick",PICK
	testl TS,TS
	__ENTER
	js 1f
	movl (PS,TS,4),TS
	__NEXT
1:	negl TS
	movl TOS(1),ecx
	js 3f
	movl ecx,4(PS,TS,4)
2:	spop
3:	spop
	__NEXT
/*;*/
//; aroll	( xx nn ..m.. mm +n adr -- xx ..m.. mm nn )
//;		( xx nn ..m.. mm -n adr -- xx mm nn ..m.. )
//;	roll within array/datafield the bottom of which is at adr
	__CODE,"aroll",AROLL
	movl TS,eax
	spop
	jmp ROLLn
/*;*/
//; -rot   	( w1 w2 w3 -- w3 w1 w2 )
//;	Rot top item to 3rd.
	__CODE,"-rot",MROT
	spush
	movl $-2,TS
	jmp jROLL
/*;*/
//; rot		( w1 w2 w3 -- w2 w3 w1 )
//;	Rot 3rd item to top.
	__CODE,"rot",ROT
	spush
	movl $2,TS
/*;*/
//; roll   	( ..w(n)..w(0) +n -- ..w(n+1)..w(0) wn )
//; roll   	( ..w(n)..w(0) -n -- ..w(0) w(n)..w(1) )
//;	roll the nth stack item to tos.
	__CODE,"roll",ROLL
	movl PS,eax	// ; data 'top', bottom up
ROLLn:	testl TS,TS	// ; count
	movl TS,ecx
	jz 2f		// ;?; nothing, drop index
	pushl edi
	pushl esi
	js 1f		// ;?; roll bot. up
	lea (eax,ecx,4),edi // ; (cmove>)
	std
	lea -4(edi),esi	// ; sp@ cell-
	movl (edi),edx	// ; >rr pick r>
	rep movsl	// ; copy top down
	cld
	movl edx,(eax)
3:
	popl esi
	popl edi
2:
	__ENTER
	popl TS	// ; drop cells count
	__NEXT
1://; -ve
	negl ecx
	movl eax,edi
	movl (eax),edx
	leal 4(edi),esi
	rep movsl
	mov edx,(edi)
	jmp 3b
/*;*/
//; cmove  	( b1 b2 u -- )
//;	non-destructively(!) copy +u bytes from b1 to b2.
	__CODE,"cmove",CMOVE
	mov TS,ecx
	testl TS,TS
	__ENTER
	movl edi,TS
	movl esi,edx
	popl edi
	popl esi
	jng 1f
	cmpl edi,esi
	jc 2f		// ; top down if src > dst
	jz 1f		// ; noop if src=dst
	rep movsb
1:
	movl TS,edi
	movl edx,esi
	popl TS
	__NEXT
2:
	lea -1(edi,ecx),edi
	lea -1(esi,ecx),esi
	std
	rep movsb
	cld
	jmp 1b
/*;*/
/*; erase  	( b u -- )*/
/*;	Fill u bytes of <nul> characters to area beginning at b.*/
	__CODE,"erase",ERASE
	spush
	xor TS,TS
	/* -> FILL */
/*;*/
/*; fill   	( b u c -- )*/
/*;	Fill u bytes of character c to area beginning at b.*/
	__CODE,"fill",FILL
	movl edi,edx
	__ENTER
	popl ecx
	popl edi
	testl ecx,ecx
	mov TS,eax
	jng 1f		/* ;prevent -ve & zero count*/
	rep stosb
1:
	popl TS
	movl edx,edi
	__NEXT
/*;*/
//; 2+		( a -- a )		F
//;	Add 2 to address.
	__CODE,"2+",TWOP
	push $2
/*;*/
//; +		( w w -- sum )		F
//;	Add top two items.
__CODE,"+",PLUS
	popl	eax
	__ENTER
	addl	eax,TS
	__NEXT
/*;*/
//; 0=		( w -- t )
//;	Ret true if top item is zero.
__CODE,"0=",ZEQ
	push $0
	/* -continued- -> EQUAL*/
/*;*/
/*; =		( w w -- t )*/
/*;	Ret true if top two are equal.*/
__CODE,"=",EQUAL
	popl eax
	__ENTER
	cmpl eax,TS
	movl $0,TS
	jz strue
	__NEXT
/*;*/
/*; <		( n1 n2 -- t )*/
/*;	Signed compare of top two items.*/
__CODE,"<",LESS
	popl %eax
	__ENTER
	cmpl TS,%eax
	movl $0,TS
	jge sfalse
strue:
	decl TS
sfalse:
	__NEXT
/*;*/
//; 1+		( a -- a )		F (Cy undeterminate)
//; char+  	( a -- a )	(n.n.)
//;	Add 1 (char size) to address.
__CODE,"1+",ONEP
	__ENTER
	incl TS
	__NEXT
/*;*/
//; 1-		( a -- a )		F (Cy undeterminate)
//; char-  	( a -- a )	(n.n.)
//;	Subtract 1 (char size) from address.
__CODE,"1-",ONEM
	__ENTER
	decl TS
	__NEXT
/*;*/
//; 4+		( a -- a )
//; cell+  	( a -- a )	(n.n.)
//;	Add cell size, 4 bytes, to address.
__CODE,"4+",CELLP
	__ENTER
	leal CELSZ(TS),TS
	__NEXT
/*;*/
//; 4-		( a -- a )
//; cell-  	( a -- a )	(n.n.)
//;	Subtract cell size, 4 bytes, from address.
__CODE,"4-",CELLM
	__ENTER
	leal -CELSZ(TS),TS
	__NEXT
/*;*/
//; shift	( n1 n -- n3 )
//;	shift n1 by +n2 bits left, -n2 bits unsigned right.
__CODE,"shift",SHIFT
	movl TS,ecx
	testl TS,TS
	__ENTER
	popl TS
	js .r
	shll cl,TS
	__NEXT
.r:
	negl ecx
	shrl cl,TS
	__NEXT
/*;*/
//; 2*		( a -- a )		F(shift opr)
//;	Multiply tos by 2
__CODE,"2*",TWOMUL
	__ENTER
	sall $1,TS
	__NEXT
/*;*/
//; 4*		( a -- a )		F(shift opr)
//; cells  	( n -- n )	(n.n.)
//;	Multiply tos by cell size in bytes, mul by 4
__CODE,"4*",CELLS
	__ENTER
	sall $2,TS
	__NEXT
/*;*/
//; 2/		( a -- a )		F(ashift opr)
//;	floored(!) divide tos by 2
__CODE,"2/",TWOBY
	__ENTER
	sarl $1,TS
	__NEXT
/*;*/
//; 4/		( a -- a )		F(ashift opr)
//; cell/  	( n -- n )	(n.n.)
//;	floored(!) divide tos by cell size in bytes, div by 4.
__CODE,"4/",CELSZBY
	__ENTER
	sarl $2,TS
	__NEXT
/*;*/
//; flg  	( -- n )
//;	Ret cpu flags from previous arithm. opr, re 'F' glossary note
//;	should be used compiled into a word, otherwise meaningless result
//;	e.g, addition w. carry flag ": +c + flg 10 and ;"
__CODE,"flg",FLG
	spush
	__ENTER
	pushfl
	pop TS
	__NEXT
/*;*/

/*;*/
/*; zcount  	( a -- a +n )*/
/*;	Ret count of an asciz string and address*/
__CODE,"zcount",ZCOUNT
	movl TS,ecx
	__ENTER
	pushl TS
	notl ecx
zcount.n:
	cmpb $1,(TS)
	incl TS
	jnc zcount.n
	addl ecx,TS
	__NEXT
/*;*/
/*; st@  	( -- flg )*/
/*;	ret content of the {state} uvari, tf while compiling*/
__CODE,"st@",STATEF
	spush
	__ENTER
	movl uSTATE(US),TS
	__NEXT
/*;*/	
/*; ---*/
/*;*/

/*;*/
/*; dovoc  	( -- )			C*/
/*;	Run time action of VOCABULARY.*/
__CODE CPLO,"dovoc",DOVOC
	leal 2*CELSZ(IP),ecx
	rpop IP
	movl ecx,uCNTXT(US)
	__NEXT
/*;*/
/*; forth  	( -- )*/
/*;	Make FORTH the context vocabulary.*/
__COLON,"forth",FORTH
	.long DOVOC
	.long 0		// ;bottom word-link
	.long LASTN	// <= ;vocabulary header, ptr is 'wid'
FORTHv:	.long 0		// ;voc-link
/*;*/
/*; ?dup     	( w -- w w | 0 )*/
/*;	Dup tos if its is not zero.*/
__CODE,"?dup",QDUP
	test TS,TS
	__ENTER
	jz 1f
	spush
1:
	__NEXT
/*;*/
/*; not		( w -- w )		F*/
/*;	One-s complement of tos.*/
__CODE,"not",INVER
	__ENTER
	notl TS
	__NEXT
/*;*/
/*; negate  	( n -- -n )		F*/
/*;	Two-s complement of tos.*/
__CODE,"negate",NEGAT
	__ENTER
	neg TS
	__NEXT
/*;*/
/*; dnegate	( d -- -d )		F*/
/*;	Twos complement of top double.*/
__CODE,"dnegate",DNEGA
	negl (esp)
	adcl $0,TS
	__ENTER
	neg TS
	__NEXT
/*;*/
//; 2-		( a -- a )		F
//;	Subtract 1 (char size) from address.
__CODE,"2-",TWOM
	__ENTER
	subl $2,TS
	__NEXT
/*;*/
/*;-		( n1 n2 -- n3 )		F*/
/*;	Subtraction.*/
__CODE,"-",SUBB
	movl TS,eax
	__ENTER
	popl TS
	subl eax,TS
	__NEXT
/*;*/
/*; abs		( n -- n )		F(sign flag)*/
/*;	Ret the absolute value of n.*/
__CODE,"abs",ABSS
	testl TS,TS
	__ENTER
	jns 1f
	negl TS
1:
	__NEXT
/*;*/
/*; u<		( u1 u2 -- f )		*/
/*;	Unsigned compare of top two items.*/
__CODE,"u<",ULESS
	movl TS,eax	// ; u2
	popl ecx	// ; u1
	__ENTER
	xorl ecx,eax
	js 1f
	subl TS,ecx	// ; u1 u2 -
	movl ecx,TS
1:
	sarl $31,TS	// ; 0<
	__NEXT
/*;*/
/*; max		( n1 n2 -- n )*/
/*;	Ret the greater of two top stack items.*/
__CODE,"max",MAX
	popl eax
	__ENTER
	cmpl eax,TS
	jng 2f
	__NEXT
/*;*/
/*; min		( n1 n2 -- n )*/
/*;	Ret the smaller of top two stack items.*/
__CODE,"min",MIN
	popl eax
	__ENTER
	cmpl eax,TS
	jl 1f
2:
	movl eax,TS
1:
	__NEXT
/*;*/
/*; within  	( u ul uh -- t )*/
/*;	Ret true if u is in range [ul,uh}.*/
__CODE,"within",WITHI
	popl eax	// ; ul	;n2
	popl ecx	// ; u	;n1
	subl eax,TS	// ; uh ul -
	subl eax,ecx	// ; u  ul -
	__ENTER
	cmpl TS,ecx	// ; (u-ul)|(uh-ul)
	sbbl TS,TS
	__NEXT
/*;*/
/*; Division*/
/*;*/

/*;*/
//; sm/rem  	( d n -- r q )
//;	Signed 'round near zero' divide double by sing. Ret mod & quotient.
//;	div by zero: ret max(sgn(dividend)), ret (1,0) if dividend = 0
//;	if quot would overflow ret unchanged dividend, for further processing.
__CODE,"sm/rem",SMBYR
	popl edx	// ; ddh
	popl eax	// ; ddl
	__ENTER
	test TS,TS	// ; dvs
	movl TS,ecx
	jz RDZ1		// ;?; div0
	cmpl $-1,TS
	jz RDM		// ;?; ovf
	idiv TS
1:
	movl eax,TS
	pushl edx
	__NEXT
RDM:	/*; div -1 */
	negl eax
	xorl edx,edx
	jmp 1b

/*;*/
//; um/mod  	( ud un -- ur uq )
//;	Unsigned divide double by sing. Ret mod & quotient.
//;	div by zero: ret max(sgn(dividend)), ret (1,0) if dividend = 0
//;	if quot would overflow ret unchanged dividend, for further processing.
__CODE,"um/mod",UMMOD
	popl edx	// ; ddh
	popl eax	// ; ddl
	__ENTER
	testl TS,TS	// ; dvs
	jz RDZ1		// ;?; div0
	cmp edx,TS
	jc 1f		// ;?; prevent ovf
	div TS
	pushl edx
	movl eax,TS
	__NEXT
1:	/* ;ret dividend from div ovf*/
	pushl eax	// ; ddl
	mov edx,TS	// ; ddh
	__NEXT
/*;*/
//; fm/mod  	( d n -- r q )
//;	Signed floored divide double by single. Ret mod & quotient.
//;	div by zero: ret max(sgn(dividend)), ret (1,0) if dividend = 0
//;	if quot would overflow ret unchanged dividend, for further processing
	__CODE,"fm/mod",FMMOD	/*:HIER: signal handler verwenden ::*/
	popl edx	// ; ddh
	popl eax	// ; ddl
	__ENTER
	testl TS,TS	// ; dvs
	movl TS,ecx	// ;		; dup 0< dup >r
	jz RDZ1		// ;?; div0
	jns 1f		// ;		; if
	negl TS		// ;		; negate >r dnegate r> endif
	negl eax
	adcl $0,edx
	negl edx
1:
	testl edx,edx
	jns 2f
	addl TS,edx
2:
	divl TS
	testl ecx,ecx
	jg 3f
	jnz 4f
	testl edx,edx
	jz 4f
	negl eax
4:
	negl edx
3:
	movl eax,TS
	pushl edx
	__NEXT
RDZ1:
	call RDZ
	jmp 3b
RDZ:	/*;dvs 0*/
	orl edx,eax
	lea 1(eax),eax
	jz RDZ3
	leal 0x80000001(,edx,2),eax
RDZ2:	/*;dvs -1*/
	negl eax
RDZ3:
	xorl edx,edx
	ret
/*;*/
/*; /mod   	( n1 n2 -- r q )*/
/*;	Signed divide. Ret mod and quotient*/
/*;	0/0: ret 1,0; n1/0 ret sgn(n1)*max,0*/
	__CODE,"/mod",SLMOD
	movl (esp),edx
	call dv
	movl edx,(esp)
	movl eax,TS
	__NEXT
/*;*/
/*; mod		( n1 n2 -- r )*/
/*;	Signed divide. Ret mod only.*/
	__CODE,"mod",MODD
	popl edx
	call dv
	movl edx,TS
	__NEXT
/*;*/
/*; /		( n1 n2 -- q )*/
/*;	Signed divide. Ret quotient only.*/
	__CODE,"/",SLASH
	popl edx
	call dv
	movl eax,TS
	__NEXT
dv:
	movl edx,eax
	sarl $31,edx
	testl TS,TS
	__ENTER
	jz RDZ
	cmpl $-1,TS
	jz RDZ2
	idiv TS
	ret
/*;*/
/*; Multiply*/
/*;*/

/*;*/
/*; um*		( u u -- ud )*/
/*;	Unsigned multiply. Ret double product.*/
	__CODE,"um*",UMSTA
	movl (esp),eax
	__ENTER
	mull TS
	movl eax,(esp)
	movl edx,TS
	__NEXT
/*;*/
/*; *		( n n -- n )*/
/*;	Signed multiply. Ret single product.*/
	__CODE,"*",STAR
	popl eax
	__ENTER
	imull TS
	movl eax,TS
	__NEXT
/*;*/
/*; m*		( n n -- d )*/
/*;	Signed multiply. Ret double product.*/
	__CODE,"m*",MSTAR
	movl TS,eax
	__ENTER
	imull (esp)
	movl eax,(esp)
	movl edx,TS
	__NEXT
/*;*/
//; */mod  	( n1 n2 n3 -- r q )
//;	Multiply n1 and n2, then divide by n3. Ret mod and quotient. floored!
	__COLON,"*/mod",SSMOD
	.long TOR,MSTAR,RFROM,SMBYR,EXIT
/*;*/
//; */		( n1 n2 n3 -- q )
//;	Multiply n1 by n2, then divide by n3. Ret quotient only. floored!
	__COLON,"*/",STASL
	.long SSMOD,SWAP,DROP,EXIT
/*;*/
/*; Miscellaneous*/
/*;*/

/*;*/
/*; aligned	( b -- a )*/
/*;	Align address to the cell boundary.*/
	__CODE,"aligned",ALGND
	addl $CELSZ-1,TS
	__ENTER
	andl $CELLMASK,TS
	__NEXT
/*;*/
/*; bl		( -- c )*/
/*;	Ret ascii <bl>, the blank character.*/
	__CODE,"bl",BLANK
	spush
	__ENTER
	movl $BL,TS
	__NEXT
/*;*/
//; key>char	( c -- c f )
//;	tf if c is printing char, c unchanged
	__CODE,"key>char",K2C
	movzbl bl,ecx
	__ENTER
	cmpb $HT,bl
	movl $pchmask,edx
	movl $-1,TS
	jz 3f		// ;?; <tab>
1:
	cmpb dl,cl
	jb 2f		// ;?; invalid below lower bound
	cmp dh,cl
	jbe 3f		// ;?; valid at & below upper bound
	shrl $16,edx
	jnz 1b		// ;?; try 2nd range
2:
	incl TS
3:
	pushl ecx
	__NEXT
/*;*/
//; >char  	( c1 -- c2 )
//;	Filter non-printing (long) characters.
//;	replaces non-printing chars by "."
	__CODE,">char",TCHAR
    .if _mbchar_
	movl TS,eax	// ; byte-cnt or zero
	__ENTER
	bsrl TS,eax	// ; find top bit to send
	leal 7(eax),eax	// ; round up to 8bit-byte size
	shrl $3,eax	// ; bytewise
	jnz 1f		// ;?; non-zero cnt
	incl eax	// ; at least one byte
1:
	leal (,eax,8),ecx
    .else
	__ENTER
    .endif
2:
	movl $pchmask,edx
3:
	cmpb dl,bl
	jb 4f		// ;?; invalid below lower bound
	cmpb dh,bl
	jbe 5f		// ;?; valid at & below upper bound
	shrl $16,edx
	jnz 3b		// ;?; try 2nd range
4:
	movb $'.',bl	// ; not a valid char, force NZ, C already cleared
5:
    .if _mbchar_
	rorl $8,TS
	decl eax
	jg 2b
	roll cl,TS
    .endif
	__NEXT
/*;*/
/*; depth  	( -- n )*/
/*;	Ret the depth of the data stack.*/
	__CODE,"depth",DEPTH
	spush
	__ENTER
	movl uSZERO(US),TS
	subl PS,TS
	sarl $2,TS
	__NEXT
/*;*/
/*; Memory access*/
/*;*/

/*;*/
//; perform	( a -- )
//;	Execute non-zero vector stored in address a
	__CODE,"perform",ATEXE
	movl (TS),eax
	testl eax,eax
	jz PSTORx
	popl TS
	jmp *eax
/*;*/
/*; +!		( n a -- )*/
/*;	Add n to the contents at address a.*/
	__CODE,"+!",PSTOR
	popl eax
	addl eax,(TS)
PSTORx:
	__ENTER
	popl TS
	__NEXT
/*;*/
//; d!		( d a -- )
//;	Store the double integer d to address a, m.s. cell to m.s. address..
	__CODE,"d!",DSTOR
	popl ecx
	popl eax
DSTOR2:
	movl TS,edx
	__ENTER
	popl TS
	movl eax,(edx)
	movl ecx,4(edx)
	__NEXT
/*;*/
//; d@		( a -- d )
//;	Fetch double integer d from address a, m.s. cell from m.s. address.
	__CODE,"d@",DFETCH
	movl (TS),eax
	movl 4(TS),TS
DFETCH2:
	__ENTER
	pushl eax
	__NEXT
/*;*/
//; 2!		( d a -- )
//;	Store double cell d to address a; re "d@" for litle endian mode.
//;	NOTE: ANS-4th inconsistency, "big-endian", m.s. cell at lower address!
	__CODE,"2!",STOR2
	popl eax
	popl ecx
	jmp DSTOR2
/*;*/
//; 2@		( a -- d )
//;	Fetch double cell d from address a; re "d@" for litle endian mode.
//;	NOTE: ANS-4th inconsistency, "big-endian", m.s. cell at lower address!
	__CODE,"2@",FETCH2
	movl 4(TS),eax
	movl (TS),TS
	jmp DFETCH2
///	__ENTER
///	pushl eax
///	__NEXT
/*;*/
/*; count  	( a1 -- a2 +n )*/
/*;	Ret count byte of a string and add 1 to byte address.*/
	__CODE,"count",COUNT
	movzbl (TS),eax
	incl TS
	__ENTER
	pushl TS
	movl eax,TS
	__NEXT	
/*;*/
/*; here   	( -- a )*/
/*;	Ret the top of the code dictionary.*/
	__CODE,"here",HERE
	spush
	__ENTER
	movl uCP(US),TS
	__NEXT
/*;*/
/*; pad		( -- a )*/
/*;	Ret the address of the text buffer above the code dictionary.*/
	__CODE,"pad",PAD
	spush
	__ENTER
	movl uCP(US),TS
	addl $PADSZ,TS
	__NEXT	
/*;*/
/*; tib		( -- a )*/
/*;	Ret the address of the terminal input buffer.*/
	__CODE,"tib",TIB
	spush
	__ENTER
	movl uTIBP(US),TS
	__NEXT
/*;*/
/*; -trailing	( b u1 -- b u2 )*/
/*;	Adjust the count to eliminate trailing white space. u1 truncated to < 64K*/
	__CODE,"-trailing",DTRAI
	mov (esp),edx
	movzwl bx,ecx
	__ENTER
	lea (ecx,edx),TS
1:
	decl TS
	decl ecx
	js 2f
	cmpb $' ',(TS)
	jz 1b
	incl TS
2:
	subl edx,TS
	__NEXT
/*;*/
//; packs  	( b u a -- a )
//;	Build counted string at aligned a with u characters from b. Null fill.
//;	-re- "place"
	__CODE,,PACKSZ		// ; for "word", build <nul>-terminated stg
	movl (esp),ecx
	movl $0,(TS,ecx)
//
	__CODE,"packs",PACKS
	lea CELSZ-1(TS),ecx	// ; a+3
	movl 4(esp),eax		// ; b
	movl (esp),TS		// ; u
	andl $CELLMASK,ecx	// ; aa
	movl TS,edx
	negl edx
	sarl $31,edx
	andl edx,TS		// ; 0 max
	movl eax,(esp)		// ; ( aa b
	movl ecx,4(esp)		// ;
	leal (TS,ecx),eax
	andl $CELLMASK,eax	// ; align to equal or next lower multiple of 4
	movl $0,(eax)		// ; 0 aa u + !
	mov bl,(ecx)		// ; u aa c! 1+ swap
	incl ecx
	pushl ecx		// ; ( aa b aa1 u -- )
	jmp CMOVE
/*;*/
/*; Numeric output, *single* precision*/
/*;*/

/*;*/
//; bs@		( -- n )
//;	fetch base value, { base @ }
	__CODE,"bs@",BASEF
	spush
	__ENTER
	movl uBASE(US),TS
	__NEXT
/*;*/
/*; digit  	( u -- c )*/
/*;	Convert digit u to a character.*/
	__CODE,"digit",DIGIT
	movzbl bl,TS
	jmp EXTRCn
/*;*/
/*; (#)		( n base -- n c )*/
/*;	Extract the least significant digit from n.*/
	__CODE,"(#)",EXTRC
	movl (esp),eax
	xorl edx,edx
	divl TS
	movl eax,(esp)
	movzbl dl,TS
EXTRCn:
	__ENTER
	cmp $10,bl
	jc 2f
	addb $7,bl
2:
	addb $'0',bl
	__NEXT
/*;*/
/*; <#		( -- )*/
/*;	Initiate the numeric output process.*/
	__CODE,"<#",LTSHARP
	movl uCP(US),eax
	addl $PADSZ,eax
	__ENTER
	movl eax,uHLD(US)
	__NEXT
/*;*/
/*; sign   	( n -- )*/
/*;	Add a minus sign to the numeric output string.*/
	__CODE,"sign",SIGN
	testl TS,TS
	movl $'-',TS
	jns DROP
	/* -continued- -> HOLD */
/*;*/
/*; hold   	( c -- )*/
/*;	Insert a character into the numeric output string.*/
	__CODE,"hold",HOLD
	movl uHLD(US),ecx
	__ENTER
	decl ecx
	movl ecx,uHLD(US)
	movb bl,(ecx)
	popl TS
	__NEXT
/*;*/
/*; #		( u -- u )*/
/*;	Extract one digit from u and append the digit to output string.*/
	__COLON,"#",SHARP
	.long	BASEF,EXTRC,HOLD,EXIT
/*;*/
/*; #s		( u -- 0 )*/
/*;	Convert u until all digits are added to the output string.*/
	__COLON,"#s",SHARPS
1:	.long	SHARP,DUP
	.long	NZBRA,1b
	.long	EXIT
/*;*/
/*; #>		( w -- b u )*/
/*;	Prepare the output string to be TYPE-d.*/
	__COLON,"#>",SHARPGT
	.long	DROP,HLD,AT
	.long	PAD,OVERM,EXIT
/*;*/
/*; <#s>   	( w -- b u )*/
/*;	Convert integer to signed numeric string.*/
	__COLON,"<#s>",SDIGS
	.long	TORR,ABSS
	.long	LTSHARP,SHARPS,RFROM
	.long	SIGN,SHARPGT,EXIT
/*;*/
/*; hex		( -- )*/
/*;	Use radix 16 as base for numeric conversions.*/
	__CODE,"hex",HEX
	__ENTER
	movl $16,uBASE(US)
	__NEXT
/*;*/
/*; decimal	( -- )*/
/*;	Use radix 10 as base for numeric conversions.*/
	__CODE,"decimal",DECIM
	__ENTER
	movl $10,uBASE(US)
	__NEXT
/*;*/
/*; Numeric input, *single* precision*/
/*;*/

/*;*/
/*; digit?  	( c base -- u f )	*/
/*;	Convert char c to numeric value u. tf indicates success.*/
	__CODE,"digit?",DIGTQ
	__ENTER
	movl (esp),ecx
	movzbl cl,eax	// ;char
	mov al,ah	// ;save char
	cmpb $'a',al
	jc 2f		// ;?; upper case low ascii
	xor $32,al
2:
	subb $'0',al	// ;#48 $30
	jc 1f		// ;?;not a valid char
	cmpb $10,al
	jc 3f		// ;?;decimal digit
	cmpb $17,al
	jc 1f		// ;?;ascii gap
	subb $7,al	// ;#55 $35
3:
	cmp bl,al
	jnc 1f		// ;?;not in range
	movzbl al,ecx	// ;char value
	decb bh		// ;true
1:
	movl ecx,(esp)
	movsbl bh,TS
	__NEXT
/*; prepare number stg conversion w/ base praefix & sign*/
	__CODE,,QN1		// ; ( a -- a 0 p u|0 sg bs )
	pushl TS		// ; ( a . a 
	pushl $0		// ; ( a 0 . a 
	__ENTER
	incl TS
	pushl TS		// ; ( a 0 p . a 
	movzbl -1(TS),ecx
	pushl ecx		// ; ( a 0 p u . a 
	movb (TS),cl		// ; 1st char
	movl $16,edx		// ; temp. base = sedecimal
	cmpb $'$',cl
	movl uBASE(US),eax	// ; base@
	jz 1f			// ; sedecimal
	cmpb $'#',cl
	jz 2f			// ; decimal
	cmpb $'@',cl
	jz 3f			// ; octal
	cmpb $'&',cl
	jz 3f			// ; octal
	cmpb $'#',cl
	jz 2f			// ; decimal
	cmpb $'%',cl
	jnz 4f			// ; use current base setting
	subl $6,edx		// ;binary
3:
	subl $2,edx		// ;octal
2:
	subl $6,edx		// ;decimal
1:
	movl edx,eax
	incl 4(esp)		// ; ( a 0 p  u  . a 
	decl (esp)		// ; ( a 0 p1 u1 . a - adjust stg ptr & count
4:
	movl eax,edx		// ; edx := temp. base
	xorl ecx,ecx		// ; +ve
	cmpb $'-',(TS)
	jnz 5f
	incl 4(esp)
	decl (esp)		// ; ( a 0 p2 u2 . a - adjust stg ptr & count
	decl ecx		// ; -ve
5:
	movl (esp),TS
	decl (esp)
	pushl TS		// ; ( a 0 p1 u1 u1
	pushl ecx		// ; ( a 0 p1 u1 u1 sg . a 
	movl uBASE(US),TS	// ; base@
	movl edx,uBASE(US)	// ; ( a 0 p1 u1 u1 sg . bs
	__NEXT
/*;*/
/*; number?	( a -- n T | a F )*/
/*;	Convert a number string to sing integer. Push a flag on tos.*/
/*;	use temporary base value correspondingto 1st char:*/
/*;		"%" binary, "&" or "@" octal, "#" decimal, "$" sedecimal*/
	__COLON,"number?",NUMBQ
	.long QN1,TOR,TOR
	.long ZBRA,6f				// ;if(6)				{ a 0 ps |u -- }b s
	.long TOR				// ;	1- 
2:	.long TORR,CAT,BASEF,DIGTQ		// ;FOR(2) >rr c@ base @ digit?
	.long ZBRA,4f				// ;if(4)
	.long SWAP,BASEF,STAR,PLUS,RFROM	// ;	swap base@ * + r>
	.long ONEP				// ;	1+
	.long DONXT,2b				// ;NEXT(2)
	.long RAT,SWAP,DROP			// ;	r@ swap drop
	.long ZBRA,3f				// ;if(3)
	.long NEGAT				// ;	negate
3:	.long SWAP				// ;n(3)	swap
	.long BRAN,5f				// ;e(5)->
4:	.long RDROP,RDROP,DDROP,ZERO		// ;n(4)	r>drop drop 2drop 0
5:	.long DUP				// ;n(5)	dup
6:	.long RDROP,DROP			// ;n(6)	r>drop drop		{ a 0 |ps -- }bs -sg-
	.long RFROM,BASE,STORE,EXIT		// ;	r> base ! ;s
/*;*/
/*; Basic I/O*/
/*;*/

/*;*/
//; sys		ccc( a -- a er )		I
//;	get/cpl literal syscall no. n by name ccc. re "(s)".
//;	e.g, send a signal:
//;		33 sp@ sys geteuid swap drop sp@ sys kill
	__COLON IMED,"sys",SYS
	.long BLW,PSYSP,LITERAL,MCPL,SC,EXIT
/*;*/
/*; ?key   	( -- c T | F )*/
/*;	Ret input character and true, or a false if no input.*/
	__CODE,"?key",QKEY
	jmp *uTQKEY(US)
/*;*/
/*; cr		( -- )*/
/*;	Output a carriage return and a line feed.*/
	__CODE,"cr",CR4
	spush
	movl $LF,TS
/*;*/
/*; emit   	( c -- )*/
/*;	Send a character to the output device; vectored to "'emit"*/
	__CODE,"emit",EMIT
	jmp *uTEMIT(US)
/*;*/
/*; space   	( -- )*/
/*;	Send a blank character to the output device.*/
	__CODE,"space",SPACE
	spush
	xorl TS,TS
	incl TS
/*;*/
/*; spaces  	( +n -- )*/
/*;	Send n spaces to the output device.*/
	__CODE,"spaces",SPACES
	xorl edx,edx
	testl TS,TS
	__ENTER
	jng 1f
	lea ' '(edx),eax
	pushal
	movl TS,esi
	incl edx
	movl uSTDOUT(US),TS
	lea s_eax(esp),ecx
2:
	sys_ write
	decl esi
	jg 2b
	popal
1:
	spop
	__NEXT
/*;*/
//; pt	( a -- )
//;	count type text at address a, "print"
	__CODE,"pt",PRINT
	movzbl (TS),eax
	incl TS
	spush
	movl eax,TS
/*;*/
/*; type   	( b +u -- )*/
/*;	Output u, at most 64K, characters from b if u > 0 and a =/= 0*/
	__CODE,"type",TYPES
	testl TS,TS
	movzwl bx,edx	// ; u, limit to 64K
	__ENTER
	popl ecx
	jng 1f
	testl ecx,ecx
	pushal
	jz 2f
	movl uSTDOUT(US),TS
	sys_ write
2:
	popal
1:
	popl TS
	__NEXT
/*;*/
/*; .r		( n +n -- )*/
/*;	Display an integer in a field of n columns, right justified.*/
	__COLON,".r",DOTR
	.long TOR,SDIGS
DOTR1:	.long RFROM
	.long OVERM
	.long SPACES,TYPES,EXIT
/*;*/
/*; u.r		( u +n -- )*/
/*;	Display an unsigned integer in n column, right justified.*/
	__COLON,"u.r",UDOTR
	.long TOR,LTSHARP,SHARPS,SHARPGT
	.long BRAN,DOTR1
/*;*/
/*; u.		( u -- )*/
/*;	Display an unsigned integer in free format, followed by a space.*/
	__COLON,"u.",UDOT
	.long LTSHARP,SHARPS,SHARPGT
	.long TYPES,SPACE,EXIT
/*;*/
/*; ?		( a -- )*/
/*;	Display the contents in a memory cell.*/
	__CODE,"?",QUEST
	movl (TS),TS
	/* -continues- -> DOT */
/*;*/
/*; .		( w -- )*/
/*;	signed Display an integer in free format, followed by a space.*/
	__COLON,".",DOT
	.long	SDIGS,TYPES,SPACE,EXIT
/*;*/
/*; Parsing*/
/*;*/

/*
; in: eax=char, ebx=stg1, ecx=count1
;out: eax=stg2, ebx=count2
;mod: a,b,c,f
*/skip:
	test ecx,ecx
	jng 1f
	pushl edi
	movl TS,edi
	cmpl $(BL<<8|BL),eax
	jz 2f
	repz scasb
	incl ecx
3:
	decl ecx
4:
	leal 1(ecx),TS	// ; rmg count
	leal -1(edi),eax // ; 1st =/= char
	popl edi
	ret
1:
	movl TS,eax	// ; a`:=a
	mov ecx,TS	// ; u`:=0
	ret
2:
	scasb		// ; skip any ctrl and <bl>
	jc 3b		// ;?; higher code
	decl ecx	// ; bl or lower
	jnz 2b
	jmp 4b		// ;>; no other chars
/*
; in: eax=char, ebx=stg1, ecx=count1
;out: eax=stg2, ebx=count2
;mod: a,b,c,f
*/scan:
	testl ecx,ecx
	jng 1b
	pushl edi
	movl TS,edi
	cmp $(BL<<8|BL),eax
	jz 5f
	repnz scasb
	jz 4b
	jmp 6f		// ; end of stg & zero if char not found
5:
	scasb		// ; skip any ctrl and <bl>
	jnc 3b #	// ;?; char above <bl>
	decl ecx	// ; bl or lower code
	jnz 5b
6:
	incl edi
	jmp 3b		// ;>; no other chars
/*;*/
//; skip   	( a +n c -- a' u' )
//;	skip leading chars = c of length u string(a,n).
//;	ret posn p' of 1st char =/= c and remaining cnt u'.
//;	scan string(a,n) for 1st occurrence of char =/= c.
//;	specific handling of <bl> including low ctrl-s if c = 0000<bl><bl>
	__CODE,"skip",SKIP
	movl TS,eax	// ;eax:=c
	popl ecx	// ;ecx:=u
	movl (PS),TS	// ;ebx:=a
	__ENTER
	call skip
	mov eax,(PS)	// ;eax=a', ebx=u'
	__NEXT
/*;*/
//; scan   	( a +n c -- a' u' )
//;	letter-case dependently scan string(a,n) for 1st occurrence of char = c.
//;	ret posn p' of 1st char = c and remaining cnt u'.
//;	ret u' = 0 and ptr to after stg if dlm not found.
//;	skip leading chars =/= c of string(a,n).
//;	specific handling of <bl> including low ctrl-s if c = 0000<bl><bl>
// ;c: ecx,edx
	__CODE,"scan",SCAN
	movl TS,eax	// ;eax:=c
	popl ecx	// ;ecx:=u
	movl (PS),TS	// ;ebx:=a
	__ENTER
	call scan
	movl eax,(PS)
	__NEXT
//
	__CODE,,PARSE41
	spush
	movl $')',TS
	/* -> PARSE*/
/*;*/
//; parse   	( c -- b u )
//;	Scan input stream(tib+@in,#tib-@in), ret counted string delimited by c.
	__CODE,"parse",PARSE
	movzbl bl,eax	// ;a:= char
	movl uNTIB(US),ecx
	movl uTOIN(US),edx
	movl uTIBP(US),TS
	__ENTER
	addl edx,TS	// ;b:=a @tib @>in +
	subl edx,ecx	// ;c:=u #tib @>in -
// ; in: eax=c, ebx=a, ecx=u1
// ;out: eax=u2, ebx=d, ecx=a, edx=a1, f=xx
	cmpb $BL,al
	pushl TS	// ; a
	jnz 1f
	movb al,ah
	call skip
	movl TS,ecx	// ; rmg count
	movl eax,TS	// ; rmg stg
	movl $(BL<<8|BL),eax // ; char|char
1:
	pushl TS	// ; a1
	call scan	// ;a:=a2(1st==c), ebx:=count2
	popl edx	// ;d:=a1
	testl TS,TS
	popl ecx	// ;c:=a
	movl eax,TS	// ;b:=a2 or a2+1
	jz 2f		// ;?; char not found
	incl TS
2:
	subl edx,eax	// ; u2   := a1-a2
	subl ecx,TS	// ; disp := a2-a
	addl TS,uTOIN(US)
	pushl edx	// ; a1
	movl eax,TS
	__NEXT
/*;*/
/*; .(		( -- )			I*/
/*;	Output following string up to next ) .*/
	__COLON IMED,".(",DOTPR
	.long PARSE41,TYPES,EXIT
/*;*/
/*; (		( -- )			I*/
/*;	Ignore following string till next ")" in same line or eol.*/
	__COLON IMED,"(",PAREN
	.long PARSE41,DDROP,EXIT
/*;*/
/*; \		( -- )			I*/
/*;	Ignore following text till the end of line.*/
	__CODE IMED,"\\",BKSLA
	movl uNTIB(US),eax
	__ENTER
	movl eax,uTOIN(US)
	__NEXT
/*;*/
/*; char   	( -- c )		I*/
/*;	Parse next word and return its first character as a literal.*/
	__COLON IMED,"char",CHAR
	.long	BLANK,PARSE,DROP,CAT,LITERAL,EXIT
/*;*/
/*; token   	( -- a	 <string> )*/
/*;	Parse a word from input stream and copy it to name dictionary.*/
	__COLON,"token",TOKEN
	.long	BLANK,PARSE
	.long	DOLIT,31,MIN
	.long	NP,AT,OVERM,CELLM
	.long	PACKS
	.long	EXIT

BLW:	/*; from "sys" *//*4*/
	pushl TS
	movl $' ',TS
/*;*/
/*; word   	( c -- a   <string> )*/
/*;	Parse a word from input stream and copy it to code dictionary, <nul> terminated.*/
	__COLON,"word",WORDD
	.long	PARSE,HERE,PACKSZ,EXIT
/*;*/
/*; Dictionary search*/
/*;*/

/*;*/
/*; name>   	( na -- ca )*/
/*;	Ret a code address given a name address.*/
	__CODE,"name>",NAMETO
	__ENTER
	movl -8(ebx),ebx
	__NEXT
/*;*/
/*; same?   	( a1 a2 u -- a1 a2 f \ -0+ )*/
/*;	Compare u cells in two strings.*/
/*;	Ret 0 if identical content, -ve if 1st stg code lower than 2nd.*/
	__CODE,"same?",SAMEQ
	testl ebx,ebx
	__ENTER
	jz 1f
	pushl edi
	pushl esi
	movl ebx,ecx
	movl 8(esp),esi		// ; 2nd
	movl 12(esp),edi
	xorl ebx,ebx
	repz cmpsl
	jz 2f
	incl ebx
	movl -4(esi),eax	// ; 2nd
	cmpl -4(edi),eax
	jc 2f
	negl ebx
2:
	popl esi
	popl edi
1:
	__NEXT
/*;*/
/*; (find)   	( a va -- ca na | a F )*/
/*;	Search a vocabulary for a string. Ret ca and na if succeeded.*/
	__COLON,"(find)",FIND
	.long	SWAP,DUP,CAT
	.long	CELSZBY,TEMP,STORE
	.long	DUP,AT,TOR,CELLP,SWAP

FIND1:	.long	AT,DUP
	.long	ZBRA,FIND6
	.long	DUP,AT
	.long	DOLIT,MASKK,ANDD,RAT
	.long	XORR
	.long	NZBRA,FIND01
	.long	CELLP,TEMP,AT,SAMEQ
	.long	ZBRA,FIND5
	.long	CELLM
FIND01:	.long	CELLM
	.long	BRAN,FIND1

FIND6:	.long	RDROP
	.long	SWAP,CELLM,SWAP,EXIT

FIND5:	.long	RDROP,SWAP,DROP
	.long	CELLM
	.long	DUP,NAMETO,SWAP,EXIT
/*;*/
//; name?   	( a -- ca na | a F )
//;	Search all stacked vocabularies for a word name,
//;	excluding @context (1st) if 2nd, equal item present.
//;	If found ret ca code address and ptr na to name address
//;	else stg address a and false. -re- "find".
	__COLON,"name?",NAMEQ
	.long	CNTXT,DUP,FETCH2
	.long	EQBRA,1f
	.long	CELLM
1:	.long	TOR
2:	.long	RFROM,CELLP,TORR
	.long	AT,DUP
	.long	ZBRA,3f
	.long	FIND,QDUP
	.long	ZBRA,2b
3:	.long	RDROP,EXIT
/*;*/
//; find   	( a -- ca 0|1|-1 )
//;	ret flag whether word in stg at a found and immediate (1).
//;	(more standard like alternative to "name?")
	__COLON,"find",FFIND
	.long NAMEQ
	.LONG 1f
1:
	testl TS,TS
	jz 2f
	testl $IMED,(TS)
	movl $1,TS
	jnz 2f
	negl TS
2:
	jmp EXIT
/*;*/
/*; Terminal response*/
/*;*/

/* ;*/
/* ; ^h		( bot eot cur -- bot eot cur )*/
/* ;	Backup the cursor by one character.*/
	__CODE,,BKSP
	cmpl 4(esp),TS
	jbe NOOP
	decl TS
	pushl TS
	movl $((BSP<<16)|(' '<<8)|BSP),TS
	jmp EMIT
/* ;*/
// ; tap		( bot eot cur c -- bot eot cur )
// ;	Accept and echo the key stroke and advance cursor.
// ;	{ over c! 1+ }
	__CODE,,TAP
	popl ecx
	__ENTER
	movb bl,(ecx)
	leal 1(ecx),TS
	__NEXT
/* ;*/
// ; kt   	( bot eot cur c -- bot eot cur | bot cur cur)
// ;	Process a control-key, <tab>, <cr>, <bs>. dft in u-vector 'tap.
	__CODE,,KTAP
	cmpb $LF,bl		// ; dup <lf> =/=
	jnz 2f			// ; if
	movl (PS),ecx		// ; emit swap drop dup ;s
	movl ecx,CELSZ(PS)
	jmp EMIT
// ;
	__CODE,,FTAP
2:	cmp $DEL,bl		// ;  <del> =/=
	jz 1f			// ;  if
	mov $BL,bl		// ;  <bl>
	jmp TAP
1:				// ;  endif
	spop			// ;  drop bs ;s
	jmp BKSP
/*;*/
/*; accept   	( b u -- b u )*/
/*;	Accept characters to input buffer. Ret with actual count.*/
	__COLON,"accept",ACCEP
	.long OVER,PLUS,OVER	// ; over + over 	; b u e u
1:	.long DDUP		// ; 2dup =/=
	.long EQBRA,4f		// ; while
	.long KEY,K2C		// ;  key key>char
	.long ZBRA,2f		// ;  if
	.long FILEQ,NZBRA,5f	// ;[hp] no echo if stdin is file
	.long DUP,EMIT		// ;[hp] echo!
5:	.long TAP
	.long BRAN,1b		// ;  else
2:	.long TTAP,ATEXE	// ;  ''tap @execute (xec "kTap", for <cr>)
3:	.long BRAN,1b		// ; endif repeat
4:	.long DROP,OVERM,EXIT	// ; drop over- ;s
/*;*/
/*; expect   	( b u -- )*/
/*;	Accept up to 164 chars from input stream and store count in "span".*/
	__COLON,"expect",EXPEC
	.long TEXPE,ATEXE,SPAN,STORE,DROP,EXIT
/*;*/
/*; query   	( -- )*/
/*;	Accept input stream to terminal input buffer.*/
/*;	store count to "#tib",  zero to ">in"*/
.if 1
	__CODE,"query",QUERY
	spush
	pushl uTIBP(US)
	pushl $QRYSZ
	movl uTEXPE(US),TS
	call DOLST
	.long EXECU
	.long 1f
1:
	movl TS,uNTIB(US)
	movl $0,uTOIN(US)
	spop 2
	jmp EXIT
.else
	__COLON,"query",QUERY
	.long TIB,DOLIT,QRYSZ,TEXPE,ATEXE
	.long NTIB,STORE
	.long DROP,ZERO,TOIN,STORE,EXIT
.endif
/*;*/
/*; Error handling*/
/*;*/

/*;*/
/*; catch   	( ca -- 0 | err# )*/
/*;	Execute word at ca and set up an error frame for it.*/
	__COLON,"catch",CATCH
	.long SPAT,TOR,HANDL,AT,TOR	/*; save error frame*/
	.long RPAT,HANDL,STORE,EXECU	/*; execute*/
	.long RFROM,HANDL,STORE		/*;restore error frame*/
	.long RDROP,ZERO,EXIT		/*; no error*/
/*;*/
//; throw   	( err# -- -?- )
//;	Reset system to current local error frame, update error flag.
// ;	(NOTE: active, lo-level replacement uncertain? needs testing)
	__CODE,"throw",THROW
	testl TS,TS		////
	movl uHANDL(US),eax
	jz .r			////
	test eax,eax
	jz 1f
	movl (eax),ecx
	movl 4(eax),PS
	movl 8(eax),IP
	leal 12(eax),RS
	movl ecx,uHANDL(US)
1:
	popl TS
	__NEXT
/*;*/
//; abort   	( -- )
//;	Reset data stack and jump to "quit"
	__CODE,"abort",ABORT
	movl $vNULLS,TS
	jmp THROW

// ;	Run time routine of ABORT" . Abort with a message
	__COLON CPLO,,ABORQ
	.long ZBRA,ABOR1	// ;text flag
	.long DOSTR,THROW	// ;pass error string 
ABOR1:	.long DOSTR,DROP,EXIT	// ;drop error

/*;*/
/*; The text interpreter*/
/*;*/

/*;*/
/*; (itp)   	( a -- )*/
/*;	Interpret a word. If failed, try to convert it to an integer.*/
	__COLON,"(itp)",INTER
	.long NAMEQ,QDUP
	.long ZBRA,INTE1
	.long AT,DOLIT,CPLO,ANDD
	.long ABORQ
	.byte 1f-.-1
	.ascii " compile only"
1:
	d_align
	.long EXECU,EXIT	// ;execute word
INTE1:	.long TNUMB,ATEXE	// ;convert number
	.long ZBRA,INTE2
	.long EXIT
INTE2:	.long THROW		// ;error
/*;*/
/*; [		( -- )			I*/
/*;	Start the text interpreter, clr {state} uvari.*/
	__CODE IMED,"[",LBRAC
	__ENTER
	movl $INTER,uTEVAL(US)
	andl $0,uSTATE(US)
	__NEXT
/*;*/
/*; .ok		( -- )*/
/*;	commandline prompt: xec .s or, "ok>"/">" -re- (ok), [ok].*/
	__COLON,".ok",DOTOK
	.long	CR4,POKP,ATEXE,EXIT
/*;*/
/*; (ok)	( -- a )*/
/*;	xec-vec of commandline prompt, dft ".s", "[ok]" at 2nd cell*/
	__VARI,"(ok)",POKP
	.long DOTS,BOKB
/*;*/
/*;	[ok]	( -- )*/
/*;	commandline prompt, "ok>" while interpreting, else ">"*/
	__CODE,"[ok]",BOKB
	spush
	cmpl $INTER,uTEVAL(US)
	movl $(('>'<<8|'k')<<8)|'o',TS//;"ok>"
	jz 1f
	shrl $16,TS			//;">" cpl
1:
	jmp EMIT
/*;*/
/*; ?stack   	( -- )*/
/*;	Abort if the data stack underflows.*/
	__COLON,"?stack",QSTAC
	.long	DEPTH,ZLESS	/*; check only for underflow*/
	.long	ABORQ
	.byte	1f-.-1
	.ascii	" underflow"
1:
	d_align
	.long	EXIT
/*;*/
/*; eval   	( -- )*/
/*;	Interpret the input stream.*/
	__COLON,"eval",EVAL
	.long BRAN,EVAL1
EVAL2:	.long TEVAL,ATEXE,QSTAC		// ; evaluate input, check stack
EVAL1:	.long TOKEN,DUP,CAT		// ;?;input stream empty
	.long NZBRA,EVAL2
	.long DROP,TPROM,ATEXE,EXIT	/*; prompt*/
/*;*/
/*; Shell*/
/*;*/

/*;*/
/*; preset   	( -- )*/
/*;	Reset data stack pointer and the terminal input buffer.*/
	__CODE,"preset",PRESET
	movl (brkaddr),eax
	movl uSZERO(US),esp
	leal -TIBB(eax),ecx	/*; terminal input buffer*/
	__ENTER
	movl ecx,uTIBP(US)
	popl TS
	__NEXT
/*;*/
/*; file   	( -- )*/
/*;	Select I/O vectors for file download.*/
	__CODE,"file",FILE
	movl $NOOP,uTPROM(US)
	movl $DROP,uTECHO(US)
	movl $FTAP,uTTAP(US)	/*; non-echoing <nl>*/
FIS:
	__ENTER
	movl $0,(sgf)
	movl $ACCEP,uTEXPE(US)
	__NEXT
/*;*/
/*; terminal   	( -- )*/
/*;	Select I/O vectors for terminal interface.*/
	__CODE,"terminal",HAND
	movl $DOTOK,uTPROM(US)
	movl $EMIT,uTECHO(US)
	movl $KTAP,uTTAP(US)	/*; echoing <nl>*/
	jmp FIS
/*;*/
/*; console   	( -- )*/
/*;	Initiate terminal interface.*/
	__CODE,"console",CONSOLE
	movl vISLO+4,ecx
	movl vISLO,edx
	movl ecx,uTEMIT(US)
	movl edx,uTQKEY(US)
	jmp HAND
/*;*/
//; fi?		( -- f )
//;	true flag if input from file
	__CODE,"fi?",FILEQ
	pushl TS
	__ENTER
	movl $0,TS
	pushal
	movl $1024,edx
	movl esp,ebp
	leal -1024(esp),ecx
	movl $self0t,ebx
	movl ecx,esp
	sys_ readlink		// ;b:=name,c:=buf,d:=bufsize
	testl eax,eax
	movl $ttyt,esi
	movl esp,edi
	jng 2f
	cmpsl
	jnz 3f
	cmpsl
	jz 2f
// ;AS ist vollends krank, noch idiotischer als NASM...
.equ ptyt,((('y<<8)|'t)<<8|'p)<<8|'/
.equ ptst,((('s<<8)|'t)<<8|'p)<<8|'/
	cmpl $ptyt,-4(edi)
	jz 2f
	cmpl $ptst,-4(edi)
	jz 2f
3:
	decl s_TS(ebp)
2:
	movl ebp,esp
	popal
1:
	__NEXT

/*;*/
/*; i/o		( -- a )*/
/*;	Array to store default I/O vectors.*/
	__VARI,"i/o",ISLO
	.long	QRX,TXSTO		/*;default I/O vectors*/
/*;*/
/*; quit   	( -- )*/
/*;	Reset return stack pointer and start text interpreter.*/
	__COLON,"quit",QUIT
	.long	RZERO,AT,RPSTO		/*;reset return stack pointer*/
QUIT1:	.long	LBRAC			/*;start interpretation*/
QUIT2:	.long	QUERY			/*;get input*/
	.long	DOLIT,EVAL,CATCH,QDUP	/*;evaluate input*/
	.long	ZBRA,QUIT2		/*;continue till error*/
/// -?-	.long	TPROM,AT,TOR		/*;*//* quatsch! s.u.*/
	.long	CONSOLE			/*;save input device*/
	.long	NULLS,OVER		/*;display error message*/
	.long	EQBRA,QUIT3
	.long	SPACE,COUNT
QUIT.c:	.long	UTYPEFF			/*;error message, truncated to max 255 chars*/
	.long	DOTQP
	.byte	1f-.-1
	.ascii	" ? "
1:
	d_align
QUIT3:
///	.long	RFROM,DOLIT,DOTOK	/*;file input*//* <= was fr ein grandioser bldsinn!*/
///	.long	EQBRA,QUIT4
///	.long	DOLIT,ESC,EMIT		/*;file error, tell host*/
QUIT4:	.long	PRESET			/*;some cleanup*/
	.long	BRAN,QUIT1
	
/*;*/
/*; The compiler*/
/*;*/

/*;*/
/*; '		( -- ca ) '*/
/*;	Search context vocabularies for the next word in input stream.*/
	__COLON,"'",TICK
	.long TOKEN,NAMEQ		/*; ?defined*/
	.long ZBRA,TICK1
	.long EXIT			/*;yes, push code address*/
TICK1:	.long THROW			/*;no, error*/
/*;*/
/*; allot   	( n -- )*/
/*;	Allocate n bytes to the code dictionary.*/
	__CODE,"allot",ALLOT
	__ENTER
	addl TS,uCP(US)
	__NEXT
/*;*/
//; 0,		( -- )
//;	Compile zero cell into the code dictionary.
	__CODE,"0,",ZCOMMA
	spush
	movl $0,TS
	/* -continued- */
/*;*/
//; ,		( w -- )
//;              Compile an integer into the code dictionary.
	__CODE,",",COMMA
	movl uCP(US),eax
	__ENTER
	movl TS,(eax)
	addl $CELSZ,eax
	popl TS
	movl eax,uCP(US)
	__NEXT
/*;*/
/*; [compile]	( --  <string> )	I*/
/*;	Compile the next immediate word into code dictionary.*/
	__COLON IMED,"[compile]",BCOMP
	.long	TICK,COMMA,EXIT
/*;*/
/*; compile	( -- )			C*/
/*;	Compile the next address in colon list to code dictionary.*/
	__CODE CPLO,"compile",CMPL
	movl 4(IP),eax		// ; r> dup @
	movl uCP(US),ecx
	addl $8,IP
	addl $CELSZ,uCP(US)	// ; , 4+ >r
	movl eax,(ecx)
	__NEXT
/*;*/
//; -compile	( -- )			C
//;	xec if interpreting, else compile next address in colon list.
	__CODE CPLO,"-compile",MCPL
	cmpl $SCOMP,uTEVAL(US)
	jz CMPL
NOOP:
	__ENTER
	__NEXT
/*;*/
//; literal	( w -- )		I
//;	in compilin state, compile tos as a literal.
	__COLON IMED,"literal",LITERAL
	.long STATEF,ZBRA,LITERx
	.long CMPL,DOLIT,COMMA
LITERx:	.long EXIT
/*;*/
/*; recurse	( -- )			I
//;	compile call to currently being defined word.
//:QUATSCH::;Make the current word available for compilation.
	__CODE IMED,"recurse",RECUR
	movl uCP(US),ecx
	movl uLAST(US),eax
	addl $CELSZ,uCP(US)
	movl -8(eax),edx
	__ENTER
	movl edx,(ecx)
	__NEXT
/*;*/
/*; Flow control abstraction threaders */
/*;*/

.if _doloop_
// ; "do" runtime code
	__CODE,"(d)",PDOP
	leal -8(RS),RS
	__ENTER
	movl TS,(RS)	// ; start
	popl eax
	popl TS
	movl eax,4(RS)	// ; limit
	__NEXT
// ; "-leave" runtime code
	__CODE,"(-)",PMLEAVEP
	testl TS,TS
	__ENTER
	spop
	jz 1f
	movl (RS),eax
	incl eax
	movl eax,4(RS)
1:
	__NEXT
// ; "+loop" runtime code
	__CODE,"(+)",PPLOOPP
	movl TS,ecx		// ; loop index
	testl TS,TS		// ; increment
	leal (ecx,TS),ecx
	popl TS
	jz PLOOPPn		// ;?; inc by one if 0
	movl 4(RS),eax
	movl (RS),ecx		// ; index
	leal 0x80000000(eax),eax
	leal 0x80000000(ecx),ecx
	js 2f			// ; -ve inc
	cmpl eax,ecx
	jnc 3f			// ; count not > limit
1:
	movl 4(IP),IP
	__NEXT
2:
	cmpl eax,ecx
	jnc 1b
PPLOOPPr:
3:
	addl $2*CELSZ,IP	// ; continue after loop
	addl $2*CELSZ,RS	// ; drop loop parm
	__NEXT
// ; "loop" runtime code
	__CODE,"(l)",PLOOPP
	mov (RS),ecx
PLOOPPn:	// ;	;<+loop zero inc.)
	incl ecx		// ; loop index
	cmpl 4(RS),ecx
	movl ecx,(RS)		// ; index
	jz PPLOOPPr
	movl 4(IP),IP
	__NEXT
/*;*/
//; -- do..loop, +loop, -leave
//;	available if enabled in Makefile or "e4a" assembly source.
/*;*/

/*;*/
//; do		( n1 n2 -- )		I
//;	pass loop paramters, n1 limit, n2 start value.
	__COLON IMED+2,"do",DO
	dd CMPL,PDOP,HERE,EXIT
/*;*/
//; -leave   	( n -- )		I
//;	if n =/= 0 early terminate a do..loop
	__COLON IMED,"-leave",MLEAVE
	dd CMPL,PMLEAVEP,EXIT
/*;*/
//; +loop   	( n -- )		I
//;	signed increment loop index by n, use 1 if n=0
//;	terminate if new index exceedes or equals loop limit
	__COLON IMED+4,"+loop",PLOOP
	dd CMPL,PPLOOPP,COMMA,EXIT
/*;*/
//; loop   	( a -- )		I
//;	increment loop index by one, terminate if top limit reached
	__COLON IMED+4,"loop",LOOP
	dd CMPL,PLOOPP,COMMA,EXIT
.endif
// ; def'd if _fornx_ set or _doloop_ unset
.if ((_fornx_)|(~_doloop_))
/*;*/
//; -- for..aft..next
//;	available if do..loop disabled in Makefile or "e4a" asm source.
//;	for..next impl. is default state, according to "eForth" model.
/*;*/

/*;*/
/*; for		( -- a )		I*/
/*;	Start a FOR-NEXT loop structure in a colon definition.*/
	__COLON IMED,"for",FOR
	.long	CMPL,TOR,HERE,EXIT
/*;*/
/*; aft		( a -- a A )		I*/
/*;	Jump to THEN in a FOR-AFT-THEN-NEXT loop the first time through.*/
	__COLON IMED,"aft",AFT
	.long	DROP,AHEAD,BEGIN,SWAP,EXIT
/*;*/
/*; next   	( a -- )		I*/
/*;	Terminate a FOR-NEXT loop structure.*/
	__COLON IMED,"next",NEXT
	.long	CMPL,DONXT,COMMA,EXIT
.endif
/*; --*/
/*;*/

/*;*/
/*; begin   	( -- a )		I*/
/*;	Start an infinite or indefinite loop structure.*/
	__CODE IMED,"begin",BEGIN
	jmp HERE
/*;*/
/*; until   	( a -- )		I*/
/*;	Terminate a BEGIN-UNTIL indefinite loop structure.*/
	__COLON IMED,"until",UNTIL
	.long	CMPL,ZBRA,COMMA,EXIT
/*;*/
/*; again   	( a -- )		I*/
/*;	Terminate a BEGIN-AGAIN infinite loop structure.*/
	__COLON IMED,"again",AGAIN
	.long	CMPL,BRAN,COMMA,EXIT
/*;*/
/*; if		( -- A )		I*/
/*;	Begin a conditional branch structure.*/
	__COLON IMED,"if",IFF
	.long	CMPL,ZBRA,HERE
	.long ZCOMMA,EXIT
/*;*/
/*; ahead   	( -- A )		I*/
/*;	Compile a forward branch instruction.*/
	__COLON IMED,"ahead",AHEAD
	.long	CMPL,BRAN,HERE
	.long ZCOMMA,EXIT
/*;*/
/*; repeat   	( A a -- )		I*/
/*;	Terminate a BEGIN-WHILE-REPEAT indefinite loop.*/
	__COLON IMED,"repeat",REPEA
	.long	AGAIN,HERE,SWAP,STORE,EXIT
/*;*/
/*; then   	( A -- )		I*/
/*; endif   	( A -- )		I*/
/*;	Terminate a conditional branch structure.*/
	__CODE IMED,"endif",ENDIF
	__CODE IMED,"then",THENN
	movl uCP(US),eax
	__ENTER
	movl eax,(TS)
	spop
	__NEXT
/*;*/
/*; else   	( A -- A )		I*/
/*;	Start the false clause in an IF-ELSE-THEN structure.*/
	__COLON IMED,"else",ELSEE
	.long	AHEAD,SWAP,THENN,EXIT
/*;*/
/*; while   	( a -- A a )		I*/
/*;	Conditional branch out of a BEGIN-WHILE-REPEAT loop.*/
	__COLON IMED,"while",WHILE
	.long	IFF,SWAP,EXIT
/*;*/
/*; abort"   	( --  <string> )	I*/
/*;	Conditional abort with an error message.*/
	__COLON IMED,"abort\042",ABRTQ
	.long	CMPL,ABORQ,STRCQ,EXIT
/*;*/
/*; - string, cpl. & interp. -*/
/*;*/

/*;*/
/*; c"		( --  <string> )	I*/
/*;	Compile/fetch an inline string literal.*/
	__CODE IMED,"c\042",STRQ
	cmpl $SCOMP,uTEVAL(US)
	jnz jQTE
	call DOLST
	.long	CMPL,STRQP,STRCQ,EXIT
/*;*/
/*; ."		( --   <string> )	CI*/
/*;	Compile an inline string literal to be typed out at run time.*/
	__COLON IMED,".\042",DOTQ
	.long STATEF,ZBRA,IDOTQ
	.long CMPL,DOTQP,STRCQ,EXIT
IDOTQ:
	.long QTE,PRINT,EXIT
/*;*/
/*; (")		( -- a )		I*/
/*;	immediately, fetch a string, don't compile*/
	__CODE IMED,"(\042)",QTE
	spush
	mov $'"',TS
	jmp WORDD
/*;*/
/*; (c")   	( -- )			C*/
/*;	Run time routine compiled by c", ret address of a cpld string.*/
	__COLON CPLO,"(c\042)",STRQP
	.long	DOSTR,EXIT
/*;*/
/*; (.")   	( -- )			C*/
/*;	Run time routine of ." - Output compiled string.*/
	__COLON CPLO,"(.\042)",DOTQP
	.long DOSTR,PRINT,EXIT
/* ;Ret address of cpld stg.*/
	__CODE,,DOSTR
	spush
	mov TOR0,TS		//; stg
	__ENTER
	movzbl (TS),edx
	leal 4(TS,edx),eax
	andl $CELLMASK,eax	//; aligned end of stg
	movl eax,TOR0		//; for next IP address
	__NEXT
/* ;Cpl literal stg up to next '"'*/
	__COLON CPLO,,STRCQ
	.long QTE,STGS
STGS:
	movzbl (TS),eax
	leal 1(TS,eax),eax
	spop
	movl $0,(eax)
	addl $CELSZ-1,eax
	andl $CELLMASK,eax
	movl eax,uCP(US)
	jmp EXIT
/*;*/
//; nul		( -- 0 )
//;	push cons zero
	__CODE,"nul",ZERO
	spush
	__ENTER
	xorl TS,TS
	__NEXT
/*;*/
/*; "nul   	( -- a ) "*/
/*;	Ret address of a null string with zero count.*/
	__VARI,"\042nul",NULLS
	.long	0
/*;*/
/*; Name compiler*/
/*;*/

/*;*/
/*; ?unique	( a -- a )*/
/*;	Display a warning message if the word already exists.*/
	__COLON,"?unique",UNIQU
	.long	DUP,NAMEQ	/*; ?name exists*/
	.long	ZBRA,UNIQ1
	.long	DOTQP		/*; redefinitions are OK*/
	.byte	1f-.-1
	.ascii	" redef "	/*; but the user should be warned*/
1:
	d_align
	.long	OVER,PRINT	/*; just in case its not planned*/
UNIQ1:	.long	DROP,EXIT
/*;*/
/*; head   	( na -- )*/
/*;	Build a new dictionary name using the string at na.*/
	__COLON,"head",SNAME
	.long DUP,CAT		/*; ?no input*/
	.long ZBRA,PNAM1
	.long UNIQU		/*; ?redef'd*/
	.long HEADS
HEADS:
	movl TS,uLAST(US)	// ;a0; dup last !
	subl $CELSZ,TS		// ; cell-
	movl uCRRNT(US),edx	// ; current @
	movl (edx),edx		// ; @
	movl edx,(TS)		// ; over !
	subl $CELSZ,TS		// ; cell-
	movl TS,uNP(US)	// ; dup np !
	movl uCP(US),eax	// ; here
	addl $CELSZ-1,eax
	andl $CELLMASK,eax	// ; aligned
	addl $3,eax		// ; 3 +
	movl eax,(TS)		// ; swap !
	spop
	jmp EXIT

PNAM1:	.long STRQP
	.byte 1f-.-1
	.ascii " name"		/*; no input*/
1:
	d_align
	.long THROW
/*;*/
/*; Forth compiler*/
/*;*/

/*;*/
/*; (cpl)   	( a -- )*/
/*;	Compile next word to code dictionary as a token or literal.*/
	__COLON 8,"(cpl)",SCOMP
.if 1
	.long	FFIND,QDUP	/*; ?defined*/
	.long	ZBRA,2f
	.long   ONEP
.else
	.long	NAMEQ,QDUP	/*; ?defined*/
	.long	ZBRA,2f
	.long	AT,DOLIT,IMED,ANDD//; ?immediate
.endif
	.long	ZBRA,1f
	.long	EXECU,EXIT	/*; its immediate, execute*/
1:	.long	COMMA,EXIT	/*; its not immediate, compile*/
2:	.long	TNUMB,ATEXE	/*; try to convert to number*/
	.long	ZBRA,3f
	.long	LITERAL,EXIT	/*; compile number as integer*/
3:	.long	THROW		/*; error*/
/*;*/
/*; overt   	( -- )*/
/*;	Link a new word into the current vocabulary.*/
	__CODE,"overt",OVERT
	movl uLAST(US),ecx
	movl uCRRNT(US),eax
	__ENTER
	movl ecx,(eax)
	__NEXT
/*;*/
/*; ;		( -- )			CI*/
/*;	Terminate a colon definition.*/
	__COLON IMED+CPLO,"\073",SEMIS
	.long	CMPL,EXIT,LBRAC,OVERT,EXIT
/*;*/
/*; ]		( -- )*/
/*;	Start compiling the words in the input stream.*/
	__CODE,"]",RBRAC
	__ENTER
	or $1,uSTATE(US)
	movl $SCOMP,uTEVAL(US)
	__NEXT
/*;*/
/*; call,	( ca -- )*/
/*;	Assemble a call instruction to ca.*/
	__CODE,"call,",CALLC
	movl uCP(US),edx
	lea 8(edx),ecx
	__ENTER
	movl ecx,uCP(US)
	movl $OCALL,(edx)
	subl ecx,TS
	movl TS,4(edx)
	popl TS
	__NEXT
/*;*/
/*; :		( --   <string> )*/
/*;	Start a new colon definition using next word as its name.*/
	__COLON,":",COLON
	.long	TOKEN,SNAME,DOLIT,DOLST
	.long	CALLC,RBRAC,EXIT
/*;*/
/*; immediate	( -- )*/
/*;	Make the last compiled word an immediate word.*/
	__CODE,"immediate",IMMED
	movl uLAST(US),eax
	__ENTER
	orb $IMED,(eax)
	__NEXT
/*;*/
/*; Defining words*/
/*;*/

/*;*/
/*; user    	( u --	 <string> )*/
/*;	Compile a new user variable.*/
	__COLON,"user",USER
	.long CREAT,CUSER
CUSER:
	movl uCP(US),eax
	addl $CELSZ,uCP(US)
	movl $DOUSE,-CELSZ(eax)
	movl TS,(eax)
	popl TS
	jmp EXIT
/*;*/
/*; create   	( --   <string> )*/
/*;	Compile a new array entry without allocating code space.*/
	__COLON,"create",CREAT
	.long	TOKEN,SNAME,OVERT
	.long	DOLIT,DOLST,CALLC
	.long	DOLIT,DOVAR,COMMA,EXIT
/*;*/
/*; variable	( --   <string> )*/
/*;	Compile a new variable initialized to 0.*/
	__COLON,"variable",VARIA
	.long CREAT
	.long ZCOMMA,EXIT
// ;
SDOES:
	movl TS,eax
	leal CELSZ(IP),TS
	movl (PS),IP
	movl eax,(PS)
	__NEXT
/*;*/
//; dds		( -- )			C
//;	cpld entry into hi-level does> part, "dodoes".
	__COLON CPLO,"dds",DDOES
	.long RFROM,LAST,AT,NAMETO,ONEP,ALGND,CELLP,STORE,EXIT
/*;*/
//; does>   	( -- addr )		I
//;	modify latest created word to jump into the DOES part 
//;	of a defining word with its pfa on the data stack.
	__COLON IMED,"does>",DOES
	.long lit(DDOES),COMMA
	.long lit(SDOES),CALLC,EXIT
/*;*/
//; constant   	ccc( n -- ) (X: -- n )
//;	initiate a named literal, a "constant"
	__COLON,"constant",CONS
	.long CREAT,CCONS
CCONS:
	movl uCP(US),eax
	movl $DOCONS,-CELSZ(eax)
	movl TS,(eax)
	popl TS
	jmp EXIT
DOCONS:
	pushl TS
	__ENTER
	movl (%esi),TS
	movl (%ebp),%esi	/*; Pop return address*/
	addl $CELSZ,%ebp	/*; Adjust RP*/
	__NEXT
/*;*/
/*; Tools*/
/*;*/

/*;*/
/*; typep   	( b u -- )*/
/*;	Display a string. Filter non-printing characters.*/
	.align 4,0x90		//; "4" => bytes, Linux/ELF specific
	nop;nop			//; align for "typep"
UTYPE31:			/*; < DOTID */
	andl $31,TS
UTYPEFF:			/*; < UTYPE */
	movzbl bl,TS
	__COLON,"typep",UTYPE
	.long	TOR		/*; start count down loop*/
	.long	BRAN,UTYP2	/*; skip first pass*/
UTYP1:
	.long	COUNT,TCHAR,EMIT
UTYP2:	.long	DONXT,UTYP1	/*; loop till done*/
	.long	DROP,EXIT
// ;	dm+	( a u -- a+u )
// ;	Dump u bytes from a, leaving a+u on the stack.
	__COLON,,DUMPP
	.long	OVER,DOLIT,4,UDOTR	/*; display address*/
	.long	SPACE,TOR		/*; start count down loop*/
	.long	BRAN,PDUM2		/*; skip first pass*/
PDUM1:	.long	DUP,CAT,DOLIT,3,UDOTR	/*; display numeric data*/
	.long	ONEP			/*; increment address*/
PDUM2:	.long	DONXT,PDUM1		/*; loop till done*/
	.long	EXIT
/*;*/
/*; dump   	( a u -- )*/
/*;	Dump u bytes from a, in a formatted manner.*/
	__COLON,"dump",DUMP
	.long BASEF,TOR,HEX		/*; save radix, set hex*/
	.long DOLIT,16,SLASH		/*; change count to lines*/
	.long TOR			/*; start count down loop*/
DUMP1:	.long CR4,DOLIT,16,DDUP
	.long DUMPP			/*; numeric*/
	.long MROT
	.long SPACE,SPACE,UTYPE		/*; printable characters*/
	.long DONXT,DUMP1		/*; loop*/
	.long BRAN,DUMP3
DUMP3:	.long DROP,RFROM,BASE,STORE	/*; restore radix*/
	.long EXIT
/*;*/
/*; .s		( ... -- ... )*/
/*;	Display the contents of the data stack.*/
	__COLON,".s",DOTS
	.long	DEPTH		/*; stack depth*/
	.long	TOR		/*; start count down loop*/
	.long	BRAN,DOTS2	/*; skip first pass*/
DOTS1:	.long	RAT,PICK,DOTSD	/*; index stack, display contents*/
DOTS2:	.long	DONXT,DOTS1	/*; loop till done*/
	.long	DOTQP
	.byte	1f-.-1
	.ascii	"<sp "
1:
	d_align
	.long	EXIT
/* from DOT - where it certainly didn`t belong to!*/
	__CODE,,DOTSD
	cmpl $10,uBASE(US)
	jz DOT
	jmp UDOT
/*;*/
/*; ?csp   	( -- )*/
/*;	Abort if stack pointer differs from that saved in CSP.*/
	__COLON,"?csp",QCSP
	.long	SPAT,CSP,AT	/*; compare pointers*/
	.long	ABORQ		/*; abort if different*/
	.byte	1f-.-1
	.ascii	"stacks"
1:
	d_align
	.long	EXIT
/*;*/
/*; !csp   	( -- )*/
/*;	Save stack pointer in CSP for error checking.*/
	__CODE,"!csp",STCSP
	__ENTER
	movl esp,uCSP(US)
	__NEXT
/*;*/
/*; >name   	( ca -- na | F )*/
/*;	Convert code address to a name address.*/
	__CODE,">name",TNAME
	leal uCNTXT(US),ecx	// ; voc-stack, 1st item
	__ENTER
1:	movl (ecx),edx		// ; @ ?dup	( ca )c:vs d:va
	testl edx,edx
	lea 4(ecx),ecx		// ; swap cell+ swap
	jz 6f
2:	// ; ebx=ca ecx=va
	movl (edx),edx		// ; @
	testl edx,edx		// ; dup
	jz 1b			// ; if
	cmpl $_start,edx
	jc 6f
	cmpl uCBRK(US),edx
	jc 5f
	cmpl uNP(US),edx
	jc 6f
	cmpl otib,edx
	jnc 6f
5:	cmpl -8(edx),TS	// ; name> =/=
	jz 3f
	subl $4,edx
	jmp 2b
6:	xorl edx,edx		// ; drop 0
3:	movl edx,TS
	__NEXT
/*;*/
/*; .id		( na -- )*/
/*;	Display the name at address.*/
	__COLON,".id",DOTID
	.long	QDUP		/*; if zero no name*/
	.long	ZBRA,DOTI1
	.long	COUNT		/*; mask lexicon bits*/
	.long	UTYPE31,EXIT	/*; display name string*/
DOTI1:	.long	DOTQP
	.byte	1f-.-1
	.ascii	" {noName}"
1:
	d_align
	.long	EXIT
/*;*/
//; see		( --    } <forth code> )
//;	A simple thread decompiler. exit on @cell = 0
	__COLON,"see",SEE
	.long TICK		/*; parse arg, get xt	*/
	.long ALGND		/*; skip a header cell	*/
	.long DUP,CELLP
SEEDOT:	.long DOT,DOLIT,10,SPACES
SEELOOP:/*;walk thread cells*/
     	.long CELLP		/*; cur'		*/
	.long DUP,AT		/*; cur, (cur)		*/
	.long ZBRA,DONESEE	/*; cur } exit if 0	*/
	/*; (cell) is non-zero*/
     	.long DUP,AT		/*; cur, (cur)		*/
	.long TNAME	    	/*; cur, nf or 0	*/
	.long DUP,ZBRA,CHECKEND/*; cur, nf or 0	*/
	.long DOTID,SPACE,SPACE	/*; cur			*/
	.long BRAN,SEELOOP
CHECKEND:/*;code header cell?	  ; cur, nf or 0	*/
	.long DROP,DUP,AT	/*; cur, (cur)		*/	
	.long DOLIT,OCALL	/*; cur, (cur), <call>	*/
	.long EQUAL,INVER	/*; cur, flag	*/
	.long ZBRA,DONESEE	/*; cur		*/	
	.long SPACE		/*; cur, 0	*/
     	.long DUP,AT,DOT	/*; cur		*/
	.long CR4,DUP
	.long BRAN,SEEDOT	/*; cur		*/
DONESEE:.long DROP,EXIT
/*;*/
/*; words   	( -- )*/
/*;	Display the names in the context vocabulary.*/
	__COLON 5,"words",WORDS
	.long	CR4,CNTXT,AT	/*; only in context*/
WORS1:	.long	AT,QDUP		/*; ?at end of list*/
	.long	ZBRA,WORS2
	.long	DUP,SPACE,DOTID	/*; display name*/
	.long	CELLM
	.long	BRAN,WORS1     
	.long	DROP
WORS2:	.long	EXIT
/*;*/
/*; System init, xec*/
/*;*/

/*;*/
//; place	( a1 a2 -- )
//;	copy counted stg from a1 to a2, add terminating <nul> byte
	__CODE,"place",PLACE
	movl TOS1,ecx		// ; stg
	movzbl (ecx),edx	// ; count
	movb dl,(TS)		// ; to mem
	movb $0,1(edx,TS)	// ; prepare terminating <nul>
	incl TS			// ; chars
	incl TOS1		// ; buf
	spush
	movl edx,TS		// ; count
	jmp CMOVE
/*;*/
//; sh		( a -- n )
//;	execute an external shell command from stg a, 
//;	wait until external command terminated, ret pid or ernum n.
//;	NOTE:	w/ linux specific "vfork" system call, 
//;		for other OS-s use "fork" -re- Makefile.
	__COLON,"sh",SYSTEM
	DD lit(CMD+3),PLACE,SHEL1
SHEL1:
	spush
	pushal
	xorl TS,TS
.if with_vfork
	sys_ vfork		// ; not recommendable, Linux specific & out-dated
.else
	sys_ fork		// ; establish 2ndary job
.endif
	testl eax,eax
	movl uCENVC(US),edx
	movl $ARGS,ecx
	jnz 4f			// ;?;continue in primary job
	movl $SHT,TS
	movl esp,ebp
	sys_ execve		// ; xec shell command, no return
4:
	writead TS,eax
	js 5f			// ;?;error from <fork>
	movl eax,TS
	xorl ecx,ecx
	xorl edx,edx
	sys_ waitpid		// ; wait until 2ndary job exitted
5:
	popal
	jmp EXIT	
/*;*/
//; ver		( -- n )
//;	Ret the 4 bytes,2 digits ea, decimal version number of this implementation.
	__CODE,"ver",VERSN
	spush
	__ENTER
	movl $VERNUM,TS
	__NEXT
/*;*/
/*; hi		( -- )*/
/*;	Initialize I/O; print system version.*/
	__COLON,"hi",HI
	.long	STOIO		/*;initialize I/O*/
	.long	CR4,DOTQP	/*;Show version*/
	.byte	1f-.-1
	.ascii	"e4a v"
	.byte	REL+48		/*;model*/
	.ascii	"."
	.byte	VER+48		/*;version*/
1:
	d_align
	.long	CR4
	.long	EXIT
/*;*/
/*; 'boot   	( -- a ) '*/
/*;	The application startup vector.*/
	__VARI,"'boot",TBOOT
	.long	STOIO		/*;application to boot*/
/*;*/
/*; cold   	( -- )*/
/*;	The hilevel cold start sequence.*/
	__COLON,"cold",COLD
COLD1:	.long DOLIT,UZERO,DOLIT,up2,AT	//;UPP
	.long DOLIT,ULAST-UZERO,CMOVE	/*;initialize user area*/
COLD3:	.long PRESET			/*;initialize data stack and TIB*/
	.long TBOOT,ATEXE		/*;application boot*/
	.long FILE
	.long FILEQ,NZBRA,COLD2
	.long CONSOLE,HI
COLD2:
	.long FORTH,CNTXT,AT,DUP	/*;initialize search order*/
	.long CRRNT,STOR2
	.long QUIT			/*;start interpretation*/

/*===============================================================*/
	__DATASEG__
CMD:	RESB 4096	//; buffer for "shell" command

LASTN	=	_NAME	/*; last name address in name dictionary*/
NTOP	=	_NTOP	/*; next available memory in name dictionary*/

	__CODESEG__

	c_align
CTOP	=	.	/*; next available memory in code dictionary*/
.end
